#!/usr/bin/env node
"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isChannelConfig = exports.isChannelConfigAuthorizedNode = void 0;
const fs_1 = __importDefault(require("fs"));
const os_1 = __importDefault(require("os"));
const yargs_1 = __importDefault(require("yargs"));
const realExternalInterface_1 = __importDefault(require("./external/real/realExternalInterface"));
const core_1 = require("./interfaces/core");
const startDaemon_1 = __importDefault(require("./startDaemon"));
// Thanks: https://stackoverflow.com/questions/4213351/make-node-js-not-exit-on-error
process.on('uncaughtException', function (err) {
    // This is important because utp-native was sporadically giving the following error and crashing:
    /*
  
    /home/root/.npm/_npx/19496/lib/node_modules/kachery-p2p-daemon/node_modules/utp-native/lib/connection.js:238
    const err = new Error(str)               ^
    Error: UTP_ECONNRESET
        at createUTPError (/home/root/.npm/_npx/19496/lib/node_modules/kachery-p2p-daemon/node_modules/utp-native/lib/connection.js:238:15)
        at Connection._onerror (/home/root/.npm/_npx/19496/lib/node_modules/kachery-p2p-daemon/node_modules/utp-native/lib/connection.js:175:16)
    Emitted 'error' event on Connection instance at:
        at Connection._onclose (/home/root/.npm/_npx/19496/lib/node_modules/kachery-p2p-daemon/node_modules/utp-native/lib/connection.js:169:25) {
      code: 'UTP_ECONNRESET',
      errno: 1
    }
  
    */
    console.warn(err.stack);
    console.log('Uncaught exception: ', err);
});
class CLIError extends Error {
    constructor(errorString) {
        super(errorString);
    }
}
const isChannelConfigAuthorizedNode = (x) => {
    return core_1._validateObject(x, {
        nodeId: core_1.isNodeId,
        nodeLabel: core_1.isNodeLabel,
        isAdmin: core_1.optional(core_1.isBoolean),
        isMessageProxy: core_1.optional(core_1.isBoolean),
        isDataProxy: core_1.optional(core_1.isBoolean),
        isPublic: core_1.optional(core_1.isBoolean)
    }, { allowAdditionalFields: true });
};
exports.isChannelConfigAuthorizedNode = isChannelConfigAuthorizedNode;
const isChannelConfig = (x) => {
    return core_1._validateObject(x, {
        channelLabel: core_1.isChannelLabel,
        bootstrapAddresses: core_1.isArrayOf(core_1.isAddress),
        authorizedNodes: core_1.isArrayOf(exports.isChannelConfigAuthorizedNode)
    }, { allowAdditionalFields: true });
};
exports.isChannelConfig = isChannelConfig;
function main() {
    const argv = yargs_1.default
        .scriptName('kachery-p2p-daemon')
        .command({
        command: 'start',
        describe: 'Start the daemon',
        builder: (y) => {
            y.option('verbose', {
                describe: 'Verbosity level.',
                type: 'number',
                default: 0
            });
            y.option('host', {
                describe: 'host name or IP address of this node.',
                type: 'string',
                default: ''
            });
            y.option('public-url', {
                describe: 'Base URL for public http access',
                type: 'string',
                default: ''
            });
            y.option('label', {
                describe: 'Label for this node (required).',
                type: 'string'
            });
            y.option('http-port', {
                describe: 'Override the default daemon http port to listen on.',
                type: 'string'
            });
            y.option('udp-port', {
                describe: 'Override the default UDP port (by default it will equal the http port).',
                type: 'string'
            });
            y.option('websocket-port', {
                describe: 'Specify the websocket port to listen on (required message proxy and/or data proxy nodes).',
                type: 'string'
            });
            y.option('isbootstrap', {
                describe: 'Mark this node as a bootstrap node',
                type: 'boolean'
            });
            y.option('noudp', {
                describe: 'Do not use a udp socket',
                type: 'boolean'
            });
            y.option('nomulticast', {
                describe: 'Do not use multicast udp',
                type: 'boolean'
            });
            y.option('static-config', {
                describe: 'Path or URL to a config file for static configuration',
                type: 'string'
            });
            y.option('auth-group', {
                describe: 'The os group that has access to this daemon',
                type: 'string'
            });
            return y;
        },
        handler: async (argv) => {
            const hostName = argv.host || null;
            const publicUrl = argv['public-url'] || null;
            const httpListenPort = argv['http-port'] ? Number(argv['http-port']) || null : null;
            const noUdp = argv['noudp'] ? true : false;
            if (noUdp) {
                if (argv['udp-port'])
                    throw Error('Cannot specify a --udp-port with the --noudp option');
            }
            const udpSocketPort = (!noUdp) ? ((argv['udp-port'] ? Number(argv['udp-port']) : httpListenPort) || 14507) : null;
            const webSocketListenPort = argv['websocket-port'] ? Number(argv['websocket-port']) : null;
            const daemonApiPort = Number(process.env.KACHERY_P2P_API_PORT || 20431);
            const label = core_1.nodeLabel(argv.label);
            const isBootstrapNode = argv['isbootstrap'] ? true : false;
            const noMulticast = argv['nomulticast'] ? true : false;
            const verbose = Number(argv.verbose || 0);
            const staticConfigPathOrUrl = argv['static-config'] ? argv['static-config'] + '' : null;
            const authGroup = argv['auth-group'] ? argv['auth-group'] + '' : null;
            const configDir = (process.env.KACHERY_P2P_CONFIG_DIR || `${os_1.default.homedir()}/.kachery-p2p`);
            // do not create the config dir because we no longer us it
            // if (!fs.existsSync(configDir.toString())) {
            //   fs.mkdirSync(configDir.toString());
            // }
            if (hostName !== null) {
                if (!core_1.isHostName(hostName)) {
                    throw new CLIError('Invalid host name');
                }
            }
            if (publicUrl !== null) {
                if (!core_1.isUrlString(publicUrl)) {
                    throw new CLIError('Invalid public url');
                }
            }
            if (!core_1.isPort(daemonApiPort)) {
                throw new CLIError(`Invalid daemon api port: ${daemonApiPort}`);
            }
            if (httpListenPort !== null) {
                if (!core_1.isPort(httpListenPort)) {
                    throw new CLIError(`Invalid http listen port: ${httpListenPort}`);
                }
            }
            if (webSocketListenPort !== null) {
                if (!core_1.isPort(webSocketListenPort)) {
                    throw new CLIError('Invalid websocket listen port');
                }
            }
            if (udpSocketPort !== null) {
                if (!core_1.isPort(udpSocketPort)) {
                    throw new CLIError('Invalid udp listen port');
                }
            }
            let storageDir = process.env['KACHERY_STORAGE_DIR'] || '';
            if (!storageDir) {
                storageDir = `${os_1.default.homedir()}/kachery-storage`;
                console.warn(`Using ${storageDir} for storage. Set KACHERY_STORAGE_DIR to override.`);
                if (!fs_1.default.existsSync(storageDir)) {
                    fs_1.default.mkdirSync(storageDir);
                }
            }
            if ((!fs_1.default.lstatSync(storageDir).isDirectory()) && (!fs_1.default.lstatSync(storageDir).isSymbolicLink)) {
                throw new CLIError(`Storage path is not a directory: ${storageDir}`);
            }
            const externalInterface = realExternalInterface_1.default(core_1.localFilePath(storageDir), configDir);
            startDaemon_1.default({
                configDir,
                verbose,
                hostName,
                publicUrl,
                daemonApiPort,
                httpListenPort,
                label,
                externalInterface,
                getDefects: () => { return {}; },
                opts: {
                    isBootstrap: isBootstrapNode,
                    multicastUdpAddress: { hostName: '237.0.0.0', port: core_1.toPort(21010) },
                    udpSocketPort,
                    webSocketListenPort,
                    firewalled: false,
                    staticConfigPathOrUrl,
                    authGroup,
                    services: {
                        announce: true,
                        discover: true,
                        bootstrap: true,
                        proxyClient: true,
                        configUpdate: true,
                        multicast: (!noUdp) ? ((!noMulticast) ? true : false) : false,
                        display: true,
                        udpSocket: !noUdp,
                        webSocketServer: webSocketListenPort ? true : false,
                        httpServer: true,
                        daemonServer: true,
                        mirror: true,
                        clientAuth: true
                    }
                }
            });
        }
    })
        .demandCommand()
        .strict()
        .help()
        .wrap(72)
        .argv;
}
main();
//# sourceMappingURL=cli.js.map