"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.JSONStringifyDeterministic = exports.createKeyPair = exports.hexToPrivateKey = exports.hexToPublicKey = exports.privateKeyToHex = exports.publicKeyHexToFeedId = exports.publicKeyToHex = exports.verifySignature = exports.verifySignatureJson = exports.getSignatureJson = exports.getSignature = void 0;
const crypto_1 = __importDefault(require("crypto"));
const core_1 = require("../interfaces/core");
const util_1 = require("./util");
const ed25519PubKeyPrefix = "302a300506032b6570032100";
const ed25519PrivateKeyPrefix = "302e020100300506032b657004220420";
const getSignature = (obj, keyPair) => {
    try {
        return crypto_1.default.sign(null, util_1.kacheryP2PSerialize(obj), keyPair.privateKey.toString()).toString('hex');
    }
    catch (err) {
        /* istanbul ignore next */
        throw Error('Exception when creating signature.');
    }
};
exports.getSignature = getSignature;
const getSignatureJson = (obj, keyPair) => {
    try {
        return crypto_1.default.sign(null, Buffer.from(exports.JSONStringifyDeterministic(obj)), keyPair.privateKey.toString()).toString('hex');
    }
    catch (err) {
        /* istanbul ignore next */
        throw Error('Exception when creating signature Json.');
    }
};
exports.getSignatureJson = getSignatureJson;
const verifySignatureJson = (obj, signature, publicKey, opts = { checkTimestamp: false }) => {
    /* istanbul ignore next */
    if (opts.checkTimestamp) {
        if (!obj.timestamp) {
            return false;
        }
        const elapsed = core_1.elapsedSince(obj.timestamp);
        // needs to be less than 30 minutes old
        const numMinutes = 30;
        if (elapsed > numMinutes * 60 * 1000) {
            return false;
        }
    }
    try {
        return crypto_1.default.verify(null, Buffer.from(exports.JSONStringifyDeterministic(obj)), publicKey.toString(), Buffer.from(signature.toString(), 'hex'));
    }
    catch (err) {
        /* istanbul ignore next */
        return false;
    }
};
exports.verifySignatureJson = verifySignatureJson;
const verifySignature = (obj, signature, publicKey, opts = { checkTimestamp: false }) => {
    /* istanbul ignore next */
    if (opts.checkTimestamp) {
        if (!obj.timestamp) {
            return false;
        }
        const elapsed = core_1.elapsedSince(obj.timestamp);
        // needs to be less than 30 minutes old
        const numMinutes = 30;
        if (elapsed > numMinutes * 60 * 1000) {
            return false;
        }
    }
    try {
        return crypto_1.default.verify(null, util_1.kacheryP2PSerialize(obj), publicKey.toString(), Buffer.from(signature.toString(), 'hex'));
    }
    catch (err) {
        /* istanbul ignore next */
        return false;
    }
};
exports.verifySignature = verifySignature;
const publicKeyToHex = (publicKey) => {
    const x = publicKey.split('\n');
    /* istanbul ignore next */
    if (x[0] !== '-----BEGIN PUBLIC KEY-----') {
        throw Error('Problem in public key format.');
    }
    /* istanbul ignore next */
    if (x[2] !== '-----END PUBLIC KEY-----') {
        throw Error('Problem in public key format.');
    }
    const ret = Buffer.from(x[1], 'base64').toString('hex');
    /* istanbul ignore next */
    if (!ret.startsWith(ed25519PubKeyPrefix)) {
        throw Error('Problem in public key format.');
    }
    return ret.slice(ed25519PubKeyPrefix.length);
};
exports.publicKeyToHex = publicKeyToHex;
const publicKeyHexToFeedId = (publicKeyHex) => {
    return publicKeyHex;
};
exports.publicKeyHexToFeedId = publicKeyHexToFeedId;
const privateKeyToHex = (privateKey) => {
    const x = privateKey.split('\n');
    /* istanbul ignore next */
    if (x[0] !== '-----BEGIN PRIVATE KEY-----') {
        throw Error('Problem in private key format.');
    }
    /* istanbul ignore next */
    if (x[2] !== '-----END PRIVATE KEY-----') {
        /* istanbul ignore next */
        throw Error('Problem in private key format.');
    }
    const ret = Buffer.from(x[1], 'base64').toString('hex');
    /* istanbul ignore next */
    if (!ret.startsWith(ed25519PrivateKeyPrefix)) {
        throw Error('Problem in private key format.');
    }
    return ret.slice(ed25519PrivateKeyPrefix.length);
};
exports.privateKeyToHex = privateKeyToHex;
const hexToPublicKey = (x) => {
    /* istanbul ignore next */
    if (!x) {
        throw Error('Error in hexToPublicKey. Input is empty.');
    }
    return `-----BEGIN PUBLIC KEY-----\n${Buffer.from(ed25519PubKeyPrefix + x, 'hex').toString('base64')}\n-----END PUBLIC KEY-----\n`;
};
exports.hexToPublicKey = hexToPublicKey;
const hexToPrivateKey = (x) => {
    /* istanbul ignore next */
    if (!x) {
        throw Error('Error in hexToPrivateKey. Input is empty.');
    }
    return `-----BEGIN PRIVATE KEY-----\n${Buffer.from(ed25519PrivateKeyPrefix + x, 'hex').toString('base64')}\n-----END PRIVATE KEY-----\n`;
};
exports.hexToPrivateKey = hexToPrivateKey;
const createKeyPair = () => {
    const { publicKey, privateKey } = crypto_1.default.generateKeyPairSync('ed25519', {
        // modulusLength: 1024,
        publicKeyEncoding: {
            type: 'spki',
            format: 'pem'
        },
        privateKeyEncoding: {
            type: 'pkcs8',
            format: 'pem'
            // cipher: 'aes-256-cbc',
            // passphrase: 'top secret'
        }
    });
    return {
        publicKey: publicKey,
        privateKey: privateKey
    };
};
exports.createKeyPair = createKeyPair;
// Thanks: https://stackoverflow.com/questions/16167581/sort-object-properties-and-json-stringify
const JSONStringifyDeterministic = (obj, space = undefined) => {
    var allKeys = [];
    JSON.stringify(obj, function (key, value) { allKeys.push(key); return value; });
    allKeys.sort();
    return JSON.stringify(obj, allKeys, space);
};
exports.JSONStringifyDeterministic = JSONStringifyDeterministic;
//# sourceMappingURL=crypto_util.js.map