"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isReadableByOthers = exports.loadYamlFromPathOrUrl = exports.loadYamlFromPath = exports.loadYamlFromUrl = exports.parseBootstrapInfo = exports.RequestTimeoutError = exports.sleepMsecNum = exports.sleepMsec = exports.kacheryP2PDeserialize = exports.kacheryP2PSerialize = exports.sha1MatchesFileKey = exports.randomAlphaString = void 0;
const assert_1 = __importDefault(require("assert"));
const axios_1 = __importDefault(require("axios"));
// somehow it doesn't work to use the default import from bson
const bson_1 = require("bson");
const fs_1 = __importDefault(require("fs"));
const js_yaml_1 = __importDefault(require("js-yaml"));
const core_1 = require("../interfaces/core");
const randomAlphaString = (num_chars) => {
    if (!num_chars) {
        /* istanbul ignore next */
        throw Error('randomAlphaString: num_chars needs to be a positive integer.');
    }
    var text = "";
    var possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
    for (var i = 0; i < num_chars; i++)
        text += possible.charAt(Math.floor(Math.random() * possible.length));
    return text;
};
exports.randomAlphaString = randomAlphaString;
const sha1MatchesFileKey = ({ sha1, fileKey }) => {
    if (fileKey.sha1) {
        return fileKey.sha1 === sha1;
    }
    return false;
};
exports.sha1MatchesFileKey = sha1MatchesFileKey;
const kacheryP2PSerialize = (x) => {
    return bson_1.serialize(sortKeysInObject(x));
};
exports.kacheryP2PSerialize = kacheryP2PSerialize;
const kacheryP2PDeserialize = (x) => {
    return convertBinaryToBufferInObject(bson_1.deserialize(x));
};
exports.kacheryP2PDeserialize = kacheryP2PDeserialize;
const sortKeysInObject = (x) => {
    if (x instanceof Buffer) {
        return x;
    }
    else if (x instanceof Object) {
        if (Array.isArray(x)) {
            return x.map(a => (sortKeysInObject(a)));
        }
        else {
            const keys = Object.keys(x).sort();
            let ret = {};
            for (let k of keys) {
                ret[k] = sortKeysInObject(x[k]);
            }
            return ret;
        }
    }
    else {
        return x;
    }
};
const convertBinaryToBufferInObject = (x) => {
    if (x instanceof bson_1.Binary) {
        // This might be the troublesome line.
        // We should check to see if the underlying type is byte array before returning the internal buffer.
        const ret = x.buffer;
        if (ret instanceof Buffer) {
            return ret;
        }
        else {
            /* istanbul ignore next */
            console.warn(ret);
            /* istanbul ignore next */
            throw Error(`Problem in convertBinaryToBufferInObject`);
        }
    }
    else if (x instanceof Object) {
        if (Array.isArray(x)) {
            return x.map(a => (convertBinaryToBufferInObject(a)));
        }
        else {
            const ret = {};
            for (let k in x) {
                ret[k] = convertBinaryToBufferInObject(x[k]);
            }
            return ret;
        }
    }
    else {
        return x;
    }
};
const sleepMsec = async (msec, continueFunction = undefined) => {
    return await exports.sleepMsecNum(msec);
};
exports.sleepMsec = sleepMsec;
const sleepMsecNum = async (msec, continueFunction = undefined) => {
    const m = msec;
    if (continueFunction) {
        const timer = core_1.nowTimestamp();
        while (m - core_1.elapsedSince(timer) > core_1.durationMsecToNumber(core_1.scaledDurationMsec(1000))) {
            if (!continueFunction()) {
                return;
            }
            await exports.sleepMsec(core_1.scaledDurationMsec(1000));
        }
        if (m > core_1.elapsedSince(timer)) {
            await exports.sleepMsec(core_1.unscaledDurationMsec(m - core_1.elapsedSince(timer)));
        }
    }
    else
        return new Promise((resolve, reject) => {
            setTimeout(() => {
                resolve();
            }, m);
        });
};
exports.sleepMsecNum = sleepMsecNum;
class StringParseError extends Error {
    constructor(errorString) {
        super(errorString);
    }
}
class RequestTimeoutError extends Error {
    constructor(errorString) {
        super(errorString);
    }
}
exports.RequestTimeoutError = RequestTimeoutError;
const parseBootstrapInfo = (x) => {
    const a = x.split(':');
    assert_1.default(a.length === 2, 'Improper bootstrap string');
    const b = {
        hostName: a[0],
        port: Number(a[1])
    };
    if (!core_1.isAddress(b)) {
        /* istanbul ignore next */
        throw new StringParseError('Improper bootstrap info.');
    }
    return b;
};
exports.parseBootstrapInfo = parseBootstrapInfo;
const cacheBust = (url) => {
    if (url.includes('?')) {
        return url + `&cb=${exports.randomAlphaString(10)}`;
    }
    else {
        return url + `?cb=${exports.randomAlphaString(10)}`;
    }
};
const loadYamlFromUrl = async (url) => {
    let txt;
    txt = (await axios_1.default.get(cacheBust(url))).data;
    return js_yaml_1.default.safeLoad(txt);
};
exports.loadYamlFromUrl = loadYamlFromUrl;
const loadYamlFromPath = async (path) => {
    let txt;
    txt = await fs_1.default.promises.readFile(path, 'utf-8');
    return js_yaml_1.default.safeLoad(txt);
};
exports.loadYamlFromPath = loadYamlFromPath;
const loadYamlFromPathOrUrl = async (pathOrUrl) => {
    if ((pathOrUrl.startsWith('http://')) || (pathOrUrl.startsWith('https://'))) {
        return await exports.loadYamlFromUrl(pathOrUrl);
    }
    else {
        return await exports.loadYamlFromPath(pathOrUrl);
    }
};
exports.loadYamlFromPathOrUrl = loadYamlFromPathOrUrl;
const isReadableByOthers = (path) => {
    const stat = fs_1.default.statSync(path);
    if ((stat.mode & fs_1.default.constants.S_IROTH) || (stat.mode & fs_1.default.constants.S_IRGRP)) {
        return true;
    }
    return false;
};
exports.isReadableByOthers = isReadableByOthers;
//# sourceMappingURL=util.js.map