"use strict";
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _currentDownloader, _bytesLoaded, _progressStream, _isRunning, _isComplete;
Object.defineProperty(exports, "__esModule", { value: true });
const DataStreamy_1 = __importDefault(require("../common/DataStreamy"));
const GarbageMap_1 = __importDefault(require("../common/GarbageMap"));
const core_1 = require("../interfaces/core");
class DownloadOptimizerJob {
    constructor(fileKey_, fileSize_, label_, findProviders, createDownloader, opts) {
        this.fileKey_ = fileKey_;
        this.fileSize_ = fileSize_;
        this.label_ = label_;
        this.findProviders = findProviders;
        this.createDownloader = createDownloader;
        this.opts = opts;
        _currentDownloader.set(this, null);
        _bytesLoaded.set(this, core_1.byteCount(0));
        _progressStream.set(this, new DataStreamy_1.default());
        _isRunning.set(this, false);
        _isComplete.set(this, false);
    }
    fileKey() {
        return this.fileKey_;
    }
    fileSize() {
        return this.fileSize_;
    }
    label() {
        return this.label_;
    }
    isDownloading() {
        return __classPrivateFieldGet(this, _currentDownloader) ? true : false;
    }
    bytesLoaded() {
        return __classPrivateFieldGet(this, _bytesLoaded);
    }
    cancel() {
        if (__classPrivateFieldGet(this, _currentDownloader)) {
            __classPrivateFieldGet(this, _currentDownloader).stop();
        }
    }
    isRunning() {
        return __classPrivateFieldGet(this, _isRunning);
    }
    isComplete() {
        return __classPrivateFieldGet(this, _isComplete);
    }
    start() {
        __classPrivateFieldSet(this, _isRunning, true);
        const timestamp = core_1.nowTimestamp();
        let numRemainingRetries = this.opts.numRetries;
        const nextTry = () => {
            let findFinished = false;
            let providerCandidates = new GarbageMap_1.default(core_1.scaledDurationMsec(60 * 60 * 1000));
            let currentDownloader = null;
            let lastDownloaderError = null;
            let somethingFound = false;
            let complete = false;
            const onFound = (providerNode) => {
                if (complete)
                    return;
                somethingFound = true;
                providerCandidates.set(providerNode.nodeId(), providerNode);
                update();
            };
            const onFindFinished = () => {
                if (complete)
                    return;
                findFinished = true;
                update();
            };
            const returnError = (err) => {
                if (complete)
                    return;
                complete = true;
                if ((numRemainingRetries <= 0) || (!somethingFound)) {
                    __classPrivateFieldSet(this, _isComplete, true);
                    __classPrivateFieldSet(this, _isRunning, false);
                    __classPrivateFieldGet(this, _progressStream).producer().error(err);
                }
                else {
                    console.info('Retrying download');
                    numRemainingRetries--;
                    setTimeout(() => {
                        nextTry();
                    }, core_1.durationMsecToNumber(core_1.scaledDurationMsec(2000)));
                }
            };
            const returnFinished = () => {
                if (complete)
                    return;
                complete = true;
                __classPrivateFieldSet(this, _isComplete, true);
                __classPrivateFieldSet(this, _isRunning, false);
                __classPrivateFieldGet(this, _progressStream).producer().end();
            };
            const handleErrorInCurrentDownloader = (err) => {
                if (!currentDownloader)
                    return;
                currentDownloader = null;
                lastDownloaderError = err;
                update();
            };
            const update = () => {
                if (complete)
                    return;
                // check if we are currently downloading something
                if (currentDownloader)
                    return;
                if (providerCandidates.values().length > 0) {
                    // we have at least one provider node... however, we might get some more so let's see if we should wait a bit more
                    const elapsed = core_1.elapsedSince(timestamp);
                    if ((!findFinished) && (elapsed < 500)) {
                        // come back in a bit
                        setTimeout(() => {
                            update();
                        }, core_1.durationMsecToNumber(core_1.scaledDurationMsec(600)));
                        return;
                    }
                    // choose the best non-busy provider node
                    const nonBusyCandidates = providerCandidates.values().filter(pn => (!pn.isDownloading()));
                    if (nonBusyCandidates.length === 0) {
                        // they are all busy, come back in a second
                        setTimeout(() => {
                            update();
                        }, core_1.durationMsecToNumber(core_1.scaledDurationMsec(1000)));
                        return;
                    }
                    const pn = chooseFastestProviderNode(nonBusyCandidates);
                    if (!pn)
                        throw Error('Unexpected, pn is null in update');
                    providerCandidates.delete(pn.nodeId());
                    currentDownloader = this.createDownloader(pn, this.fileSize(), this.label());
                    currentDownloader.start().then((ds) => {
                        ds.onError((err) => {
                            // error downloading
                            handleErrorInCurrentDownloader(err);
                        });
                        ds.onFinished(() => {
                            // we finished downloading!
                            returnFinished();
                        });
                        ds.onProgress((progress) => {
                            __classPrivateFieldGet(this, _progressStream).producer().setProgress(progress);
                        });
                    }).catch(err => {
                        // error starting the downloader
                        handleErrorInCurrentDownloader(err);
                    });
                }
                else {
                    // no provider candidates yet
                    if (findFinished) {
                        // we have finished finding, so we'll return an error
                        if (lastDownloaderError) {
                            returnError(lastDownloaderError);
                        }
                        else {
                            if (!somethingFound) {
                                returnError(Error('File not found'));
                            }
                            else {
                                returnError(Error('Problem in download'));
                            }
                        }
                        return;
                    }
                    else {
                        // we'll wait until update is called again
                    }
                }
            };
            this.findProviders(onFound, onFindFinished);
        };
        nextTry();
    }
    // setDownloader(downloader: {start: () => Promise<DataStreamy>, stop: () => void}) {
    //     if (this.#currentDownloader !== null) {
    //         /* istanbul ignore next */
    //         throw Error('Unexpected: job already has a downloader')
    //     }
    //     this.#currentDownloader = downloader
    //     downloader.start().then((ds) => {
    //         ds.onStarted((size: ByteCount | null) => {
    //             this.#progressStream.producer().start(size)
    //         })
    //         ds.onProgress((progress: DataStreamyProgress) => {
    //             this.#bytesLoaded = progress.bytesLoaded
    //             this.#progressStream.producer().setProgress(progress)
    //         });
    //         const _handleComplete = () => {
    //             this.#currentDownloader = null
    //         }
    //         ds.onError((err: Error) => {
    //             _handleComplete();
    //             this.#progressStream.producer().error(err)
    //         });
    //         ds.onFinished(() => {
    //             _handleComplete();
    //             this.#progressStream.producer().end()
    //         });
    //     }).catch((err) => {
    //         throw Error(`Error starting downloader: ${err.message}`)
    //     })
    // }
    getProgressStream() {
        return __classPrivateFieldGet(this, _progressStream);
    }
}
exports.default = DownloadOptimizerJob;
_currentDownloader = new WeakMap(), _bytesLoaded = new WeakMap(), _progressStream = new WeakMap(), _isRunning = new WeakMap(), _isComplete = new WeakMap();
const chooseFastestProviderNode = (providerNodeList) => {
    if (providerNodeList.length === 0) {
        return null;
    }
    const estimatedRates = providerNodeList.map(p => p.estimatedRateBps()).map(x => Number(x));
    const bestIndex = argMax(estimatedRates);
    return providerNodeList[bestIndex];
};
// thanks: https://gist.github.com/engelen/fbce4476c9e68c52ff7e5c2da5c24a28
function argMax(array) {
    return array.map((x, i) => [x, i]).reduce((r, a) => (a[0] > r[0] ? a : r))[1];
}
//# sourceMappingURL=DownloadOptimizerJob.js.map