"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.formatByteCount = void 0;
const constants_1 = require("../common/constants");
const DataStreamy_1 = __importDefault(require("../common/DataStreamy"));
const core_1 = require("../interfaces/core");
const NodeToNodeRequest_1 = require("../interfaces/NodeToNodeRequest");
const createDownloader = (node, fileKey, providerNode, fileSize, label) => {
    let _cancelled = false;
    const nodeId = providerNode.nodeId();
    let o = null;
    const _start = async () => {
        const timestamp = core_1.nowTimestamp();
        const ret = new DataStreamy_1.default();
        if (_cancelled) {
            ret.producer().error(Error('Cancelled'));
            return ret;
        }
        const r = await node.kacheryStorageManager().findFile(fileKey);
        if (_cancelled) {
            ret.producer().error(Error('Cancelled'));
            return ret;
        }
        if (r.found) {
            ret.producer().end();
            return ret;
        }
        const _data = [];
        const n = node.remoteNodeManager().getRemoteNode(nodeId);
        /* istanbul ignore next */
        if (!n) {
            throw Error('Unexpected. Remote node not found.');
        }
        let requestData;
        if (node.getDefects().badDownloadFileDataRequest) {
            requestData = {
                requestType: 'downloadFileData',
                fileKey,
                startByte: core_1.byteCount(10),
                endByte: core_1.byteCount(0)
            };
        }
        else {
            requestData = {
                requestType: 'downloadFileData',
                fileKey,
                startByte: core_1.byteCount(0),
                endByte: null
            };
        }
        const responseData = await n.sendRequest(requestData, { timeoutMsec: constants_1.TIMEOUTS.defaultRequest, method: 'default' });
        if (_cancelled) {
            ret.producer().error(Error('Cancelled'));
            return ret;
        }
        /* istanbul ignore next */
        if (!NodeToNodeRequest_1.isDownloadFileDataResponseData(responseData)) {
            throw Error('Unexpected response data for downloadFileData request');
        }
        if (!responseData.success) {
            ret.producer().error(Error(`Unable to stream file data: ${responseData.errorMessage}`));
            return ret;
        }
        if (!responseData.streamId) {
            throw Error('Unexpected: no stream ID');
        }
        o = await n.downloadFileData(responseData.streamId, { method: 'default' });
        o.dataStream.onError(err => {
            if (!o)
                throw Error('Unexpected in onError of createDownloader');
            const bytesLoaded = ret.bytesLoaded();
            const elapsedSec = core_1.elapsedSince(timestamp) / 1000;
            console.log(`Error downloading file data. Downloaded ${exports.formatByteCount(ret.bytesLoaded())} bytes in ${elapsedSec} sec from ${nodeId.slice(0, 6)} using ${o.method}`);
            ret.producer().error(err);
        });
        o.dataStream.onFinished(() => {
            if (!o)
                throw Error('Unexpected in onFinished of createDownloader');
            const bytesLoaded = ret.bytesLoaded();
            const elapsedSec = core_1.elapsedSince(timestamp) / 1000;
            const rate = (core_1.byteCountToNumber(bytesLoaded) / 1e6) / elapsedSec;
            console.info(`${label}: Downloaded ${exports.formatByteCount(ret.bytesLoaded())} in ${elapsedSec} sec [${rate.toFixed(3)} MiB/sec] from ${nodeId.slice(0, 6)} using ${o.method}`);
            const data = Buffer.concat(_data);
            node.kacheryStorageManager().storeFile(fileKey.sha1, data).then(() => {
                ret.producer().end();
            }).catch((err) => {
                ret.producer().error(err);
            });
        });
        o.dataStream.onStarted((size) => {
            ret.producer().start(size);
        });
        o.dataStream.onData((buf) => {
            _data.push(buf);
            ret.producer().data(buf);
        });
        ret.producer().onCancelled(() => {
            if (!o)
                throw Error('Unexpected in onCancelled of createDownloader');
            o.dataStream.cancel();
        });
        return ret;
    };
    const _stop = () => {
        if (o)
            o.dataStream.cancel();
        _cancelled = true;
    };
    return {
        start: _start,
        stop: _stop
    };
};
const formatByteCount = (n) => {
    const a = core_1.byteCountToNumber(n);
    if (a < 10000) {
        return `${a} bytes`;
    }
    else if (a < 100 * 1000) {
        return `${(a / 1000).toFixed(1)} KiB`;
    }
    else if (a < 100 * 1000 * 1000) {
        return `${(a / (1000 * 1000)).toFixed(1)} MiB`;
    }
    else {
        return `${(a / (1000 * 1000)).toFixed(0)} MiB`;
    }
};
exports.formatByteCount = formatByteCount;
exports.default = createDownloader;
//# sourceMappingURL=createDownloader.js.map