"use strict";
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _db, _initializing, _initialized, _databaseLocked, _onDatabaseLockReleasedCallbacks;
Object.defineProperty(exports, "__esModule", { value: true });
const sqlite_1 = require("sqlite");
const sqlite3_1 = __importDefault(require("sqlite3"));
const crypto_util_1 = require("../../common/crypto_util");
const util_1 = require("../../common/util");
const core_1 = require("../../interfaces/core");
const fs_1 = __importDefault(require("fs"));
class LocalFeedsDatabase {
    constructor(databasePath) {
        this.databasePath = databasePath;
        _db.set(this, null);
        _initializing.set(this, false);
        _initialized.set(this, false);
        _databaseLocked.set(this, false);
        _onDatabaseLockReleasedCallbacks.set(this, []);
    }
    async _initialize() {
        if (__classPrivateFieldGet(this, _initialized))
            return;
        while (__classPrivateFieldGet(this, _initializing)) {
            await util_1.sleepMsec(core_1.unscaledDurationMsec(100));
        }
        if (__classPrivateFieldGet(this, _initialized))
            return;
        __classPrivateFieldSet(this, _initializing, true);
        const createTables = async (db) => {
            await db.run(`
                CREATE TABLE IF NOT EXISTS feeds (
                    feedId TEXT PRIMARY KEY NOT NULL
                ) WITHOUT ROWID;
            `);
            await db.run(`
                CREATE TABLE IF NOT EXISTS subfeeds (
                    feedId TEXT NOT NULL,
                    subfeedHash TEXT NOT NULL,
                    accessRules TEXT,
                    PRIMARY KEY(feedId, subfeedHash),
                    FOREIGN KEY (feedId)
                    REFERENCES feeds (feedId) 
                        ON UPDATE CASCADE
                        ON DELETE CASCADE
                ) WITHOUT ROWID;
            `);
            await db.run(`
                CREATE TABLE IF NOT EXISTS subfeedMessages (
                    feedId TEXT NOT NULL,
                    subfeedHash TEXT NOT NULL,
                    position INTEGER NOT NULL,
                    message TEXT,
                    PRIMARY KEY(feedId, subfeedHash, position),
                    FOREIGN KEY (feedId, subfeedHash)
                        REFERENCES subfeeds (feedId, subfeedHash) 
                            ON UPDATE CASCADE
                            ON DELETE CASCADE
                ) WITHOUT ROWID;
            `);
        };
        const db = await this._openDatabase();
        try {
            await db.run(`PRAGMA foreign_keys = ON`);
            await createTables(db);
        }
        finally {
            await this._closeDatabase();
        }
        fs_1.default.chmodSync(this.databasePath.toString(), fs_1.default.constants.S_IRUSR | fs_1.default.constants.S_IWUSR);
        __classPrivateFieldSet(this, _initialized, true);
        __classPrivateFieldSet(this, _initializing, false);
    }
    async _acquireDatabaseLock() {
        if (!__classPrivateFieldGet(this, _databaseLocked)) {
            __classPrivateFieldSet(this, _databaseLocked, true);
            return;
        }
        await new Promise((resolve) => {
            __classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks).push(() => {
                __classPrivateFieldSet(this, _databaseLocked, true);
                resolve();
            });
        });
    }
    _releaseDatabaseLock() {
        if (!__classPrivateFieldGet(this, _databaseLocked)) {
            throw Error('Unexpected: cannot release database lock when it is not locked');
        }
        __classPrivateFieldSet(this, _databaseLocked, false);
        if (__classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks).length > 0) {
            const cb = __classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks)[0];
            __classPrivateFieldSet(this, _onDatabaseLockReleasedCallbacks, __classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks).slice(1));
            cb();
        }
    }
    async _openDatabase() {
        await this._acquireDatabaseLock();
        __classPrivateFieldSet(this, _db, await sqlite_1.open({ filename: this.databasePath.toString(), driver: sqlite3_1.default.Database }));
        return __classPrivateFieldGet(this, _db);
    }
    async _closeDatabase() {
        if (!__classPrivateFieldGet(this, _db))
            throw Error('Unexpected in closeDatabase: database is null');
        await __classPrivateFieldGet(this, _db).close();
        this._releaseDatabaseLock();
    }
    async addFeed(feedId) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            await db.run('BEGIN TRANSACTION');
            await db.run(`
                INSERT INTO feeds (feedId) VALUES ($feedId)
            `, {
                '$feedId': feedId.toString()
            });
            await db.run('COMMIT');
        }
        finally {
            await this._closeDatabase();
        }
    }
    async deleteFeed(feedId) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            await db.run('BEGIN TRANSACTION');
            await db.run(`
                DELETE FROM feeds WHERE feedId = $feedId
            `, {
                '$feedId': feedId.toString()
            });
            await db.run('COMMIT');
        }
        finally {
            await this._closeDatabase();
        }
    }
    async hasFeed(feedId) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            const row = await db.get(`
                SELECT feedId FROM feeds WHERE feedId = $feedId
            `, {
                '$feedId': feedId.toString()
            });
            if (row) {
                if (row.feedId === feedId.toString()) {
                    return true;
                }
                else {
                    // not expected, but checking anyway
                    throw Error('Unexpected: mismatch in feedId');
                }
            }
            else {
                return false;
            }
        }
        finally {
            await this._closeDatabase();
        }
    }
    async getSignedSubfeedMessages(feedId, subfeedHash) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            const rows = await db.all(`
                SELECT message, position FROM subfeedMessages WHERE feedId = $feedId AND subfeedHash = $subfeedHash ORDER BY position ASC
            `, {
                '$feedId': feedId.toString(),
                '$subfeedHash': subfeedHash.toString()
            });
            if (!rows) {
                throw Error('Unexpected: rows undefined');
            }
            const ret = [];
            for (let i = 0; i < rows.length; i++) {
                const row = rows[i];
                if (row.position !== i) {
                    // this enforces that feed messages are unreadable if they have gaps,
                    // but the way feed message numbers are assigned in FeedManager should keep that from ever happening.
                    throw Error(`Unexpected position in signed subfeed message: ${feedId} ${subfeedHash} ${i} <> ${row.position}`);
                }
                let m;
                try {
                    m = JSON.parse(row.message);
                }
                catch (err) {
                    throw err;
                }
                if (!core_1.isSignedSubfeedMessage(m)) {
                    throw Error(`Problem in signed subfeed message from database: ${feedId} ${subfeedHash} ${i}`);
                }
                if (m.body.messageNumber !== i) {
                    throw Error(`Unexpected message number in signed subfeed message: ${feedId} ${subfeedHash} ${i} <> ${m.body.messageNumber}`);
                }
                ret.push(m);
            }
            return ret;
        }
        finally {
            await this._closeDatabase();
        }
    }
    async getSubfeedAccessRules(feedId, subfeedHash) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            const row = await db.get(`
                SELECT accessRules FROM subfeeds WHERE feedId = $feedId AND subfeedHash = $subfeedHash
            `, {
                '$feedId': feedId.toString(),
                '$subfeedHash': subfeedHash.toString()
            });
            if (row) {
                if (!core_1.isJSONObject(row)) {
                    throw Error(`Problem getting access rules of subfeed (*): ${feedId} ${subfeedHash}`);
                }
                if (!row.accessRules)
                    return null;
                let accessRulesJson = row.accessRules;
                if (!core_1.isString(accessRulesJson)) {
                    throw Error(`Problem getting access rules of subfeed: ${feedId} ${subfeedHash}`);
                }
                let accessRules;
                try {
                    accessRules = JSON.parse(accessRulesJson);
                }
                catch (err) {
                    throw Error(`Problem parsing access rules of subfeed: ${feedId} ${subfeedHash}`);
                }
                if (!core_1.isSubfeedAccessRules(accessRules)) {
                    throw Error(`Problem in access rules of subfeed: ${feedId} ${subfeedHash}`);
                }
                return accessRules;
            }
            else {
                return null;
            }
        }
        finally {
            await this._closeDatabase();
        }
    }
    async setSubfeedAccessRules(feedId, subfeedHash, accessRules) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            await this._createFeedRowIfNeeded(db, feedId);
            await db.run('BEGIN TRANSACTION');
            await db.run(`
                INSERT OR REPLACE INTO subfeeds (feedId, subfeedHash, accessRules) VALUES ($feedId, $subfeedHash, $accessRules)
            `, {
                '$feedId': feedId.toString(),
                '$subfeedHash': subfeedHash.toString(),
                '$accessRules': crypto_util_1.JSONStringifyDeterministic(accessRules)
            });
            await db.run('COMMIT');
        }
        finally {
            await this._closeDatabase();
        }
    }
    async appendSignedMessagesToSubfeed(feedId, subfeedHash, messages) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            if (messages.length === 0)
                return;
            // CHAIN:append_messages:step(7)
            await this._createSubfeedRowIfNeeded(db, feedId, subfeedHash);
            const data = messages.map(m => ({
                feedId,
                subfeedHash,
                position: m.body.messageNumber,
                message: crypto_util_1.JSONStringifyDeterministic(m)
            }));
            await db.run('BEGIN TRANSACTION');
            for (let d of data) {
                // await fs.promises.appendFile(this.databasePath + '.debug', JSON.stringify({feedId: d.feedId, subfeedHash: d.subfeedHash, position: d.position, message: d.message}) + '\n')
                await db.run(`
                    INSERT INTO subfeedMessages (feedId, subfeedHash, position, message) VALUES ($feedId, $subfeedHash, $position, $message)
                `, {
                    '$feedId': d.feedId,
                    '$subfeedHash': d.subfeedHash,
                    '$position': d.position,
                    '$message': d.message
                });
            }
            await db.run('COMMIT');
        }
        finally {
            await this._closeDatabase();
            // CHAIN:append_messages:step(8)
        }
    }
    async _createFeedRowIfNeeded(db, feedId) {
        await db.run(`
            INSERT OR IGNORE INTO feeds (feedId) VALUES ($feedId)
        `, {
            '$feedId': feedId.toString()
        });
    }
    async _createSubfeedRowIfNeeded(db, feedId, subfeedHash) {
        await this._createFeedRowIfNeeded(db, feedId);
        await db.run(`
            INSERT OR IGNORE INTO subfeeds (feedId, subfeedHash) VALUES ($feedId, $subfeedHash)
        `, {
            '$feedId': feedId.toString(),
            '$subfeedHash': subfeedHash.toString()
        });
    }
}
_db = new WeakMap(), _initializing = new WeakMap(), _initialized = new WeakMap(), _databaseLocked = new WeakMap(), _onDatabaseLockReleasedCallbacks = new WeakMap();
exports.default = LocalFeedsDatabase;
//# sourceMappingURL=LocalFeedsDatabase.js.map