"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const crypto_util_1 = require("../../common/crypto_util");
const core_1 = require("../../interfaces/core");
class LocalFeedsDatabaseOld {
    constructor(storageDir) {
        this.storageDir = storageDir;
    }
    async addFeed(feedId) {
        // Create the feed directory for the actual feed
        await _createFeedDirectoryIfNeeded(this.storageDir, feedId); ////////////////////////////////
    }
    async deleteFeed(feedId) {
        // todo
    }
    async hasFeed(feedId) {
        return fs_1.default.existsSync(_feedDirectory(this.storageDir, feedId));
    }
    async getSignedSubfeedMessages(feedId, subfeedHash) {
        const subfeedMessagesPath = _subfeedDirectory(this.storageDir, feedId, subfeedHash) + '/messages';
        return await readMessagesFile(subfeedMessagesPath);
    }
    async getSubfeedAccessRules(feedId, subfeedHash) {
        await _createSubfeedDirectoryIfNeeded(this.storageDir, feedId, subfeedHash);
        const accessRulesPath = _subfeedDirectory(this.storageDir, feedId, subfeedHash) + '/access';
        const accessRules = await readJsonFile(accessRulesPath, { rules: [] });
        if (!core_1.isSubfeedAccessRules(accessRules)) {
            throw Error('Invalid access rules.');
        }
        return accessRules;
    }
    async setSubfeedAccessRules(feedId, subfeedHash, accessRules) {
        const subfeedDir = _subfeedDirectory(this.storageDir, feedId, subfeedHash);
        fs_1.default.mkdirSync(subfeedDir, { recursive: true });
        const accessRulesPath = subfeedDir + '/access';
        writeJsonFileSync(accessRulesPath, accessRules);
    }
    async appendSignedMessagesToSubfeed(feedId, subfeedHash, messages) {
        const textLinesToAppend = messages.map(sm => (crypto_util_1.JSONStringifyDeterministic(sm)));
        const subfeedDir = _subfeedDirectory(this.storageDir, feedId, subfeedHash);
        fs_1.default.mkdirSync(subfeedDir, { recursive: true });
        const subfeedMessagesPath = subfeedDir + '/messages';
        fs_1.default.appendFileSync(subfeedMessagesPath, textLinesToAppend.join('\n') + '\n', { encoding: 'utf8' });
    }
}
const writeJsonFileSync = (path, obj) => {
    const txt = crypto_util_1.JSONStringifyDeterministic(obj, 4);
    fs_1.default.writeFileSync(path, txt);
};
const _createFeedDirectoryIfNeeded = async (storageDir, feedId) => {
    const path = _feedDirectory(storageDir, feedId);
    if (!fs_1.default.existsSync(path)) {
        await fs_1.default.promises.mkdir(path, { recursive: true });
    }
    if (!fs_1.default.existsSync(path + '/subfeeds')) {
        await fs_1.default.promises.mkdir(path + '/subfeeds', { recursive: true });
    }
};
const _feedDirectory = (storageDir, feedId) => {
    return `${_feedParentDirectory(storageDir, feedId)}/${feedId}`;
};
const _feedParentDirectory = (storageDir, feedId) => {
    return `${storageDir}/feeds/${feedId[0]}${feedId[1]}/${feedId[2]}${feedId[3]}/${feedId[4]}${feedId[5]}`;
};
const _subfeedDirectory = (storageDir, feedId, subfeedHash) => {
    const feedDir = _feedDirectory(storageDir, feedId);
    return `${feedDir}/subfeeds/${subfeedHash[0]}${subfeedHash[1]}/${subfeedHash[2]}${subfeedHash[3]}/${subfeedHash[4]}${subfeedHash[5]}/${subfeedHash}`;
};
const readMessagesFile = async (path) => {
    let txt;
    try {
        txt = await fs_1.default.promises.readFile(path, { encoding: 'utf8' });
    }
    catch (err) {
        return [];
    }
    if (typeof (txt) !== 'string') {
        throw Error('Unexpected: txt is not a string.');
    }
    let messages = [];
    const lines = txt.split('\n');
    for (let line of lines) {
        if (line) {
            let signedMessage;
            try {
                signedMessage = JSON.parse(line);
            }
            catch (err) {
                console.warn(`Problem parsing JSON from file.`, { path });
                return [];
            }
            if (core_1.isSignedSubfeedMessage(signedMessage)) {
                messages.push(signedMessage);
            }
            else {
                console.warn(`Problem with signed message from JSON file.`, { path });
                return [];
            }
        }
    }
    return messages;
};
const _createSubfeedDirectoryIfNeeded = async (storageDir, feedId, subfeedHash) => {
    const path = _subfeedDirectory(storageDir, feedId, subfeedHash);
    if (!fs_1.default.existsSync(path)) {
        await fs_1.default.promises.mkdir(path, { recursive: true });
    }
};
const readJsonFile = async (path, defaultVal) => {
    try {
        const txt = await fs_1.default.promises.readFile(path, 'utf-8');
        return JSON.parse(txt);
    }
    catch (err) {
        return defaultVal;
    }
};
exports.default = LocalFeedsDatabaseOld;
//# sourceMappingURL=LocalFeedsDatabaseOld.js.map