"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _storageDir;
Object.defineProperty(exports, "__esModule", { value: true });
exports.renameAndCheck = exports.createTemporaryFilePath = exports.KacheryStorageManager = void 0;
const crypto_1 = __importDefault(require("crypto"));
const fs_1 = __importDefault(require("fs"));
const DataStreamy_1 = __importDefault(require("../../../common/DataStreamy"));
const util_1 = require("../../../common/util");
const core_1 = require("../../../interfaces/core");
class KacheryStorageManager {
    constructor(storageDir) {
        _storageDir.set(this, void 0);
        if (!fs_1.default.existsSync(storageDir.toString())) {
            throw Error(`Kachery storage directory does not exist: ${storageDir}`);
        }
        __classPrivateFieldSet(this, _storageDir, storageDir);
    }
    async findFile(fileKey) {
        if (fileKey.sha1) {
            const { path: filePath, size: fileSize } = await this._getLocalFileInfo(fileKey.sha1);
            if ((filePath) && (fileSize !== null)) {
                return { found: true, size: fileSize, localFilePath: filePath };
            }
        }
        if (fileKey.chunkOf) {
            const { path: filePath, size: fileSize } = await this._getLocalFileInfo(fileKey.chunkOf.fileKey.sha1);
            if (filePath) {
                const offset = fileKey.chunkOf.startByte;
                const size = core_1.byteCount(core_1.byteCountToNumber(fileKey.chunkOf.endByte) - core_1.byteCountToNumber(fileKey.chunkOf.startByte));
                return { found: true, size, localFilePath: null }; // in this case it's not the entire file, so we are not going to return the local file path
            }
        }
        return { found: false, size: core_1.byteCount(0), localFilePath: null };
    }
    async storeFile(sha1, data) {
        const s = sha1;
        const destParentPath = `${__classPrivateFieldGet(this, _storageDir)}/sha1/${s[0]}${s[1]}/${s[2]}${s[3]}/${s[4]}${s[5]}`;
        const destPath = `${destParentPath}/${s}`;
        if (fs_1.default.existsSync(destPath)) {
            return;
        }
        fs_1.default.mkdirSync(destParentPath, { recursive: true });
        const destPathTmp = `${destPath}.${util_1.randomAlphaString(5)}.tmp`;
        await fs_1.default.promises.writeFile(destPathTmp, data);
        if (fs_1.default.existsSync(destPath)) {
            /* istanbul ignore next */
            {
                fs_1.default.unlinkSync(destPathTmp);
                return;
            }
        }
        await exports.renameAndCheck(destPathTmp, destPath, data.length);
    }
    async storeFileFromStream(ds, fileSize) {
        const tmpDestPath = `${__classPrivateFieldGet(this, _storageDir)}/store.file.${util_1.randomAlphaString(10)}.tmp`;
        const writeStream = fs_1.default.createWriteStream(tmpDestPath);
        const shasum = crypto_1.default.createHash('sha1');
        const manifestChunks = [];
        const manifestData = { buffers: [], byte1: 0, byte2: 0 };
        let complete = false;
        const chunkSize = 20 * 1000 * 1000;
        const _updateManifestChunks = ({ final }) => {
            if ((manifestData.byte2 - manifestData.byte1 >= chunkSize) || ((final) && (manifestData.byte2 > manifestData.byte1))) {
                const d = Buffer.concat(manifestData.buffers);
                manifestData.buffers = [];
                for (let i = 0; i < d.length; i += chunkSize) {
                    const x = d.slice(i, i + Math.min(chunkSize, d.length - i));
                    if ((x.length === chunkSize) || (final)) {
                        manifestChunks.push({
                            start: core_1.byteCount(manifestData.byte1),
                            end: core_1.byteCount(manifestData.byte1 + x.length),
                            sha1: computeSha1OfBufferSync(x) // note that this is synchronous (not ideal)
                        });
                        manifestData.byte1 += x.length;
                    }
                    else {
                        manifestData.buffers.push(x);
                    }
                }
            }
        };
        return new Promise((resolve, reject) => {
            const _cleanup = () => {
                try {
                    fs_1.default.unlinkSync(tmpDestPath);
                }
                catch (e) {
                }
            };
            ds.onData(buf => {
                if (complete)
                    return;
                shasum.update(buf);
                writeStream.write(buf);
                manifestData.buffers.push(buf);
                manifestData.byte2 += buf.length;
                _updateManifestChunks({ final: false });
            });
            ds.onError(err => {
                if (complete)
                    return;
                complete = true;
                _cleanup();
                reject(err);
            });
            ds.onFinished(() => {
                if (complete)
                    return;
                complete = true;
                try {
                    const sha1Computed = shasum.digest('hex');
                    const s = sha1Computed;
                    const destParentPath = `${__classPrivateFieldGet(this, _storageDir)}/sha1/${s[0]}${s[1]}/${s[2]}${s[3]}/${s[4]}${s[5]}`;
                    const destPath = `${destParentPath}/${s}`;
                    const nextStep = () => {
                        _updateManifestChunks({ final: true });
                        const manifest = {
                            size: core_1.byteCount(manifestData.byte2),
                            sha1: sha1Computed,
                            chunks: manifestChunks
                        };
                        let manifestSha1 = null;
                        if (manifestChunks.length > 1) {
                            const manifestJson = Buffer.from(JSON.stringify(manifest), 'utf-8');
                            manifestSha1 = computeSha1OfBufferSync(manifestJson);
                            this.storeFile(manifestSha1, manifestJson);
                        }
                        resolve({ sha1: sha1Computed, manifestSha1 });
                    };
                    if (fs_1.default.existsSync(destPath)) {
                        // if the dest path already exists, we already have the file and we are good
                        nextStep();
                    }
                    else {
                        // dest path does not already exist
                        fs_1.default.mkdirSync(destParentPath, { recursive: true });
                        exports.renameAndCheck(tmpDestPath, destPath, core_1.byteCountToNumber(fileSize)).then(nextStep);
                    }
                }
                catch (err2) {
                    _cleanup();
                    reject(err2);
                }
            });
        });
    }
    async storeLocalFile(localFilePath) {
        let stat0;
        try {
            stat0 = await fs_1.default.promises.stat(localFilePath.toString());
        }
        catch (err) {
            throw Error(`Unable to stat file. Perhaps the kachery-p2p daemon does not have permission to read this file: ${localFilePath}`);
        }
        const fileSize = core_1.byteCount(stat0.size);
        const ds = createDataStreamForFile(localFilePath, core_1.byteCount(0), fileSize);
        return await this.storeFileFromStream(ds, fileSize);
    }
    async concatenateChunksAndStoreResult(sha1, chunkSha1s) {
        const s = sha1;
        const destParentPath = `${__classPrivateFieldGet(this, _storageDir)}/sha1/${s[0]}${s[1]}/${s[2]}${s[3]}/${s[4]}${s[5]}`;
        const destPath = `${destParentPath}/${s}`;
        if (fs_1.default.existsSync(destPath)) {
            // already exists
            /* istanbul ignore next */
            return;
        }
        // verify we have all the files
        for (let chunkSha1 of chunkSha1s) {
            const f = await this.findFile({ sha1: chunkSha1 });
            if (!f.found) {
                /* istanbul ignore next */
                throw Error(`Cannot concatenate chunk. Missing chunk: ${chunkSha1}`);
            }
        }
        const tmpPath = exports.createTemporaryFilePath({ storageDir: __classPrivateFieldGet(this, _storageDir), prefix: 'kachery-p2p-concat-' });
        const writeStream = fs_1.default.createWriteStream(tmpPath);
        const shasum = crypto_1.default.createHash('sha1');
        let totalSizeBytes = 0;
        for (let chunkSha1 of chunkSha1s) {
            const readStream = await this.getFileReadStream({ sha1: chunkSha1 });
            await new Promise((resolve, reject) => {
                readStream.onData(buf => {
                    totalSizeBytes += buf.length;
                    shasum.update(buf);
                    writeStream.write(buf);
                });
                readStream.onError(err => {
                    reject(err);
                });
                readStream.onFinished(() => {
                    resolve();
                });
            });
        }
        await new Promise((resolve, reject) => {
            writeStream.end(() => {
                const sha1Computed = shasum.digest('hex');
                if (sha1Computed !== sha1) {
                    /* istanbul ignore next */
                    {
                        /* istanbul ignore next */
                        {
                            reject(Error('Did not get the expected SHA-1 sum for concatenated file'));
                            return;
                        }
                    }
                }
                resolve();
            });
        });
        if (fs_1.default.existsSync(destPath)) {
            // already exists
            /* istanbul ignore next */
            {
                fs_1.default.unlinkSync(tmpPath);
                return;
            }
        }
        fs_1.default.mkdirSync(destParentPath, { recursive: true });
        await exports.renameAndCheck(tmpPath, destPath, totalSizeBytes);
    }
    async hasLocalFile(fileKey) {
        if (fileKey.sha1) {
            const { path: filePath, size: fileSize } = await this._getLocalFileInfo(fileKey.sha1);
            if ((filePath) && (fileSize !== null)) {
                return true;
            }
        }
        if (fileKey.chunkOf) {
            const { path: filePath, size: fileSize } = await this._getLocalFileInfo(fileKey.chunkOf.fileKey.sha1);
            if (filePath) {
                return true;
            }
        }
        return false;
    }
    async getFileReadStream(fileKey) {
        if (fileKey.sha1) {
            const { path: filePath, size: fileSize } = await this._getLocalFileInfo(fileKey.sha1);
            if ((filePath) && (fileSize !== null)) {
                return createDataStreamForFile(filePath, core_1.byteCount(0), fileSize);
            }
        }
        if (fileKey.chunkOf) {
            const { path: filePath, size: fileSize } = await this._getLocalFileInfo(fileKey.chunkOf.fileKey.sha1);
            if (filePath) {
                const offset = fileKey.chunkOf.startByte;
                const size = core_1.byteCount(core_1.byteCountToNumber(fileKey.chunkOf.endByte) - core_1.byteCountToNumber(fileKey.chunkOf.startByte));
                return createDataStreamForFile(filePath, offset, size);
            }
        }
        throw Error('Unable get data read stream for local file.');
    }
    async _getLocalFileInfo(fileSha1) {
        const s = fileSha1;
        const path = core_1.localFilePath(`${__classPrivateFieldGet(this, _storageDir)}/sha1/${s[0]}${s[1]}/${s[2]}${s[3]}/${s[4]}${s[5]}/${s}`);
        let stat0;
        try {
            stat0 = await fs_1.default.promises.stat(path.toString());
        }
        catch (err) {
            return { path: null, size: null };
        }
        return {
            path,
            size: core_1.byteCount(stat0.size)
        };
    }
    storageDir() {
        return __classPrivateFieldGet(this, _storageDir);
    }
}
exports.KacheryStorageManager = KacheryStorageManager;
_storageDir = new WeakMap();
const computeSha1OfBufferSync = (buf) => {
    const shasum = crypto_1.default.createHash('sha1');
    shasum.update(buf);
    return shasum.digest('hex');
};
const createDataStreamForFile = (path, offset, size) => {
    // note.. for some reason if we put {encoding: 'binary'} we get text data chunks
    const readStream = fs_1.default.createReadStream(path.toString(), { start: core_1.byteCountToNumber(offset), end: core_1.byteCountToNumber(offset) + core_1.byteCountToNumber(size) - 1 });
    const ret = new DataStreamy_1.default();
    ret.producer().start(size);
    readStream.on('data', (chunk) => {
        if (!core_1.isBuffer(chunk)) {
            throw Error('Unexpected type of data chunk');
        }
        ret.producer().data(chunk);
    });
    readStream.on('end', () => {
        ret.producer().end();
    });
    readStream.on('error', (err) => {
        ret.producer().error(err);
    });
    ret.producer().onCancelled(() => {
        readStream.close();
    });
    return ret;
};
const createTemporaryFilePath = (args) => {
    const dirPath = args.storageDir + '/tmp';
    fs_1.default.mkdirSync(dirPath, { recursive: true });
    return `${dirPath}/${args.prefix}-${util_1.randomAlphaString(10)}`;
};
exports.createTemporaryFilePath = createTemporaryFilePath;
const renameAndCheck = async (srcPath, dstPath, expectedSizeBytes) => {
    try {
        // this line occassionaly fails on our ceph system and it is unclear the reason. So I am catching the error to troubleshoot
        fs_1.default.renameSync(srcPath, dstPath);
    }
    catch (err) {
        if (!fs_1.default.existsSync(dstPath)) {
            throw Error(`Unexpected problem renaming file. File does not exist: ${dstPath}: ${err.message}`);
        }
        throw Error(`Unexpected problem renaming file. Even though file exists: ${dstPath}: ${err.message}`);
    }
    // we need to stat the file here for purpose of flushing to disk (problem encountered on franklab system)
    const timeoutMsec = 1000 * 10;
    const timer = core_1.nowTimestamp();
    while (true) {
        const size0 = fs_1.default.statSync(dstPath).size;
        if (size0 === expectedSizeBytes)
            break; // we are good
        await util_1.sleepMsec(core_1.scaledDurationMsec(100));
        const elapsed = core_1.elapsedSince(timer);
        if (elapsed > timeoutMsec) {
            throw Error(`Unexpected: file does not have expected size after renaming (*): ${dstPath} ${size0} ${expectedSizeBytes}`);
        }
    }
    fs_1.default.chmodSync(dstPath, fs_1.default.constants.S_IRUSR | fs_1.default.constants.S_IWUSR | fs_1.default.constants.S_IRGRP | fs_1.default.constants.S_IROTH);
};
exports.renameAndCheck = renameAndCheck;
//# sourceMappingURL=KacheryStorageManager.js.map