"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _node, _localFeedManager, _subfeeds, _remoteFeedManager, _incomingSubfeedSubscriptionManager, _outgoingSubfeedSubscriptionManager;
Object.defineProperty(exports, "__esModule", { value: true });
const GarbageMap_1 = __importDefault(require("../common/GarbageMap"));
const util_1 = require("../common/util");
const core_1 = require("../interfaces/core");
const NewIncomingSubfeedSubscriptionManager_1 = __importDefault(require("./NewIncomingSubfeedSubscriptionManager"));
const NewOutgoingSubfeedSubscriptionManager_1 = __importDefault(require("./NewOutgoingSubfeedSubscriptionManager"));
const RemoteFeedManager_1 = __importDefault(require("./RemoteFeedManager"));
const Subfeed_1 = __importDefault(require("./Subfeed"));
class FeedManager {
    constructor(node, localFeedManager) {
        // Manages the local feeds and access to the remote feeds in the p2p network
        _node.set(this, void 0); // The kachery-p2p daemon
        _localFeedManager.set(this, void 0);
        _subfeeds.set(this, new GarbageMap_1.default(core_1.scaledDurationMsec(8 * 60 * 1000))); // The subfeed instances (Subfeed()) that have been loaded into memory
        _remoteFeedManager.set(this, void 0); // Manages the interaction with feeds on remote nodes
        _incomingSubfeedSubscriptionManager.set(this, void 0);
        _outgoingSubfeedSubscriptionManager.set(this, void 0);
        __classPrivateFieldSet(this, _node, node);
        __classPrivateFieldSet(this, _localFeedManager, localFeedManager);
        __classPrivateFieldSet(this, _incomingSubfeedSubscriptionManager, new NewIncomingSubfeedSubscriptionManager_1.default(node));
        __classPrivateFieldSet(this, _outgoingSubfeedSubscriptionManager, new NewOutgoingSubfeedSubscriptionManager_1.default(node));
        __classPrivateFieldSet(this, _remoteFeedManager, new RemoteFeedManager_1.default(__classPrivateFieldGet(this, _node), __classPrivateFieldGet(this, _outgoingSubfeedSubscriptionManager)));
    }
    async createFeed({ feedName }) {
        // Create a new writeable feed on this node and return the ID of the new feed
        return await __classPrivateFieldGet(this, _localFeedManager).createFeed(feedName);
    }
    async deleteFeed({ feedId }) {
        await __classPrivateFieldGet(this, _localFeedManager).deleteFeed(feedId);
    }
    async getFeedId({ feedName }) {
        return await __classPrivateFieldGet(this, _localFeedManager).getFeedId(feedName);
    }
    async hasWriteableFeed(feedId) {
        return await __classPrivateFieldGet(this, _localFeedManager).hasWriteableFeed(feedId);
    }
    async appendMessages(args) {
        // Append messages to a subfeed (must be in a writeable feed on this node)
        // Load the subfeed and make sure it is writeable
        const subfeed = await this._loadSubfeed(args.feedId, args.subfeedHash);
        if (!subfeed) {
            /* istanbul ignore next */
            throw Error(`Unable to load subfeed: ${args.feedId} ${args.subfeedHash}`);
        }
        if (!subfeed.isWriteable()) {
            throw Error(`Subfeed is not writeable: ${args.feedId} ${args.subfeedHash}`);
        }
        const release = await subfeed.acquireLock();
        try {
            // Append the messages
            // CHAIN:append_messages:step(3)
            await subfeed.appendMessages(args.messages, { metaData: undefined });
        }
        finally {
            release();
        }
    }
    async submitMessage({ feedId, subfeedHash, message, timeoutMsec }) {
        // Same as appendMessages, except if we don't have a writeable feed, we submit it to the p2p network
        // and then, on success, it will append the messages on the node where the feed is writeable
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        if (!subfeed) {
            /* istanbul ignore next */
            throw Error(`Unable to load subfeed: ${feedId} ${subfeedHash}`);
        }
        if (subfeed.isWriteable()) {
            // If writeable, let's just append the messages
            await this.appendMessages({ feedId, subfeedHash, messages: [core_1.submittedSubfeedMessageToSubfeedMessage(message)] });
            return;
            // throw Error(`Cannot submit messages. Subfeed is writeable: ${feedId} ${subfeedHash}`);
        }
        // Submit the messages to the p2p network
        await __classPrivateFieldGet(this, _remoteFeedManager).submitMessage({ feedId, subfeedHash, message, timeoutMsec });
    }
    async getNumLocalMessages({ feedId, subfeedHash }) {
        // Get the total number of messages in the local feed only
        // future: we may want to optionally do a p2p search, and retrieve the number of messages without retrieving the actual messages
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        if (!subfeed) {
            /* istanbul ignore next */
            throw Error(`Unable to load subfeed: ${feedId} ${subfeedHash}`);
        }
        return subfeed.getNumLocalMessages();
    }
    async getFinalLocalMessage({ feedId, subfeedHash }) {
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        if (!subfeed) {
            return null;
        }
        // important to do it this way so we can load new messages into memory (if available)
        const numMessages = Number(subfeed.getNumLocalMessages());
        const position = core_1.subfeedPosition(Math.max(0, numMessages - 1));
        const messages = subfeed.getLocalSignedMessages({ position, numMessages: core_1.messageCount(1) });
        if (messages.length > 0) {
            return messages[messages.length - 1].body.message;
        }
        else
            return null;
    }
    async getFeedInfo({ feedId, timeoutMsec }) {
        // Get the p2p information about the feed
        // If this is a local and writeable, just return {isWriteable: true}
        // Otherwise we search the p2p network for the feed and if it is found we return {isWriteable: false}
        // Otherwise, if not found, throws an exception
        const privateKey = await __classPrivateFieldGet(this, _localFeedManager).getPrivateKeyForFeed(feedId);
        if (privateKey) {
            return {
                nodeId: __classPrivateFieldGet(this, _node).nodeId()
            };
        }
        else {
            // Get the liveFeedInfo. If not found, this will throw an error.
            const liveFeedInfo = await __classPrivateFieldGet(this, _remoteFeedManager).findLiveFeedInfo({ feedId, timeoutMsec });
            return liveFeedInfo;
        }
    }
    async getAccessRules({ feedId, subfeedHash }) {
        // Get the access rules for a local writeable subfeed
        // These determine which remote nodes have permission to submit messages
        // to this subfeed.
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        if (!subfeed) {
            /* istanbul ignore next */
            throw Error(`Unable to load subfeed: ${feedId} ${subfeedHash}`);
        }
        if (!subfeed.isWriteable()) {
            throw Error('Cannot get access rules for subfeed that is not writeable');
        }
        const release = await subfeed.acquireLock();
        let accessRules;
        try {
            accessRules = await subfeed.getAccessRules();
        }
        finally {
            release();
        }
        return accessRules;
    }
    async setAccessRules({ feedId, subfeedHash, accessRules }) {
        // Set the access rules for a local writeable subfeed
        // These determine which remote nodes have permission to submit messages to this subfeed
        // to this subfeed.
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        if (!subfeed) {
            /* istanbul ignore next */
            throw Error(`Unable to load subfeed: ${feedId} ${subfeedHash}`);
        }
        if (!subfeed.isWriteable()) {
            throw Error('Cannot set access rules for subfeed that is not writeable');
        }
        const release = await subfeed.acquireLock();
        try {
            await subfeed.setAccessRules(accessRules);
        }
        finally {
            release();
        }
    }
    async watchForNewMessages({ subfeedWatches, waitMsec, maxNumMessages, signed }) {
        // assert(typeof(waitMsec) === 'number');
        // assert(typeof(maxNumMessages) === 'number');
        const messages = new Map();
        // first check for messages we have locally and return those if nonempty...
        // if we have none of those, then we will wait for new messages (if waitMsec > 0)
        let foundSomething = false;
        for (let wn of subfeedWatches.keys()) {
            const watchName = wn;
            const w = subfeedWatches.get(watchName);
            if (!w)
                throw Error('Unexpected.');
            const subfeed = await this._loadSubfeed(w.feedId, w.subfeedHash);
            const numLocalMessages = subfeed.getNumLocalMessages();
            let numMessages = core_1.messageCountToNumber(numLocalMessages) - core_1.subfeedPositionToNumber(w.position);
            if (numMessages > 0) {
                foundSomething = true;
                if (core_1.messageCountToNumber(maxNumMessages) > 0) {
                    numMessages = Math.min(core_1.messageCountToNumber(maxNumMessages), numMessages);
                }
                let messages0 = subfeed.getLocalSignedMessages({ position: w.position, numMessages: core_1.messageCount(numMessages) });
                if (signed) {
                    messages.set(watchName, messages0);
                }
                else {
                    messages.set(watchName, messages0.map(m => (m.body.message)));
                }
            }
        }
        if (foundSomething) {
            return messages;
        }
        if (core_1.durationMsecToNumber(waitMsec) === 0) {
            // we are not going to wait
            return messages;
        }
        // now we are going to wait
        return new Promise((resolve, reject) => {
            // Wait until new messages are received on one or more subfeeds, and return information on which watches were triggered
            let finished = false;
            let numMessages = 0;
            const doFinish = async () => {
                if (finished)
                    return;
                if (numMessages > 0) {
                    // maybe we have other messages coming in at exactly the same time. Wait a bit for those
                    await util_1.sleepMsec(core_1.scaledDurationMsec(30));
                }
                finished = true;
                resolve(messages);
            };
            subfeedWatches.forEach((w, watchName) => {
                messages.set(watchName, []);
            });
            subfeedWatches.forEach((w, watchName) => {
                this._loadSubfeed(w.feedId, w.subfeedHash).then((subfeed) => {
                    if (subfeed) {
                        subfeed.waitForSignedMessages({ position: w.position, maxNumMessages, waitMsec }).then((messages0) => {
                            if (messages0.length > 0) {
                                if (signed) {
                                    messages.set(watchName, messages0);
                                }
                                else {
                                    messages.set(watchName, messages0.map(m => m.body.message));
                                }
                                numMessages += messages0.length;
                                if (!finished)
                                    doFinish();
                            }
                        });
                    }
                });
            });
            setTimeout(() => {
                if (!finished)
                    doFinish();
            }, core_1.durationMsecToNumber(waitMsec));
        });
    }
    async renewIncomingSubfeedSubscription(fromNodeId, feedId, subfeedHash) {
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        if (!subfeed.isWriteable()) {
            throw Error('Cannot have an incoming subscription to a subfeed that is not writeable');
        }
        // CHAIN:get_remote_messages:step(11)
        __classPrivateFieldGet(this, _incomingSubfeedSubscriptionManager).createOrRenewIncomingSubscription(fromNodeId, feedId, subfeedHash);
        return subfeed.getNumLocalMessages();
    }
    async reportNumRemoteMessages(remoteNodeId, feedId, subfeedHash, numRemoteMessages) {
        const subfeed = await this._loadSubfeed(feedId, subfeedHash);
        subfeed.reportNumRemoteMessages(remoteNodeId, numRemoteMessages);
    }
    async _loadSubfeed(feedId, subfeedHash) {
        const timer = core_1.nowTimestamp();
        // Load a subfeed (Subfeed() instance
        // If we have already loaded it into memory, then do not reload
        const k = core_1.feedSubfeedId(feedId, subfeedHash);
        let subfeed = __classPrivateFieldGet(this, _subfeeds).get(k) || null;
        if (subfeed) {
            await subfeed.waitUntilInitialized();
        }
        else {
            // Instantiate and initialize the subfeed
            subfeed = new Subfeed_1.default(__classPrivateFieldGet(this, _node), feedId, subfeedHash, __classPrivateFieldGet(this, _localFeedManager), __classPrivateFieldGet(this, _remoteFeedManager));
            subfeed.onMessagesAdded(() => {
                if (!subfeed)
                    throw Error('Unexpected');
                // CHAIN:append_messages:step(10)
                __classPrivateFieldGet(this, _incomingSubfeedSubscriptionManager).reportMessagesAdded(feedId, subfeedHash, subfeed.getNumLocalMessages() || core_1.messageCount(0));
            });
            // Store in memory for future access (the order is important here, see waitUntilInitialized above)
            __classPrivateFieldGet(this, _subfeeds).set(k, subfeed);
            // Load private key if this is writeable (otherwise, privateKey will be null)
            // important to do this after setting this.#subfeeds(k), because we need to await it
            const privateKey = await __classPrivateFieldGet(this, _localFeedManager).getPrivateKeyForFeed(feedId);
            try {
                await subfeed.initialize(privateKey);
            }
            catch (err) {
                /* istanbul ignore next */
                __classPrivateFieldGet(this, _subfeeds).delete(k);
                /* istanbul ignore next */
                throw err;
            }
        }
        // Return the subfeed instance
        return subfeed;
    }
}
_node = new WeakMap(), _localFeedManager = new WeakMap(), _subfeeds = new WeakMap(), _remoteFeedManager = new WeakMap(), _incomingSubfeedSubscriptionManager = new WeakMap(), _outgoingSubfeedSubscriptionManager = new WeakMap();
// we use this interface to ensure consistency between the in-memory signed messages and the in-database signed messages (this is crucial for integrity of p2p feed system)
// const _subfeedHash = (subfeedName) => {
//     if (typeof(subfeedName) == 'string') {
//         if (subfeedName.startsWith('~')) {
//             assert(subfeedName.length === 41, `Invalid subfeed name: ${subfeedName}`);
//             validateSha1Hash(subfeedName.slice(1));
//             return subfeedName.slice(1);
//         }
//         return sha1sum(subfeedName);
//     }
//     else {
//         return sha1sum(JSONStringifyDeterministic(subfeedName));
//     }
// }
exports.default = FeedManager;
//# sourceMappingURL=FeedManager.js.map