"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.loadFile = exports.loadFileAsync = void 0;
const DataStreamy_1 = __importDefault(require("./common/DataStreamy"));
const util_1 = require("./common/util");
const createDownloader_1 = require("./downloadOptimizer/createDownloader");
const core_1 = require("./interfaces/core");
const loadFileAsync = async (node, fileKey, opts) => {
    const r = await node.kacheryStorageManager().findFile(fileKey);
    if (r.found) {
        return r;
    }
    const ds = await exports.loadFile(node, fileKey, opts);
    return await new Promise((resolve, reject) => {
        ds.onError(err => {
            reject(err);
        });
        ds.onFinished(() => {
            node.kacheryStorageManager().findFile(fileKey).then((r) => {
                if (r.found) {
                    resolve(r);
                }
                else {
                    reject(Error('Unexpected - unable to findFile after loadFile reported finished.'));
                }
            }, (err) => {
                reject(err);
            });
        });
    });
};
exports.loadFileAsync = loadFileAsync;
async function asyncLoop(list, func, opts) {
    return new Promise((resolve, reject) => {
        let i = 0;
        let numComplete = 0;
        let numRunning = 0;
        let error = false;
        const update = () => {
            if (error)
                return;
            if (numComplete === list.length) {
                resolve();
                return;
            }
            if (numRunning < opts.numSimultaneous) {
                if (i < list.length) {
                    i++;
                    numRunning++;
                    func(list[i - 1], i - 1).then(() => {
                        numComplete++;
                        numRunning--;
                        process.nextTick(update);
                    })
                        .catch((err) => {
                        error = true;
                        numRunning--;
                        reject(err);
                    });
                }
            }
        };
        update();
    });
}
const loadFile = async (node, fileKey, opts) => {
    const { fromNode } = opts;
    const r = await node.kacheryStorageManager().findFile(fileKey);
    if (r.found) {
        const ret = new DataStreamy_1.default();
        ret.producer().end();
        return ret;
    }
    const loadFileWithManifest = async () => {
        const manifestSha1 = fileKey.manifestSha1;
        if (!manifestSha1)
            throw Error('Unexpected');
        const entireFileTimestamp = core_1.nowTimestamp();
        const ret = new DataStreamy_1.default();
        const manifestFileKey = { sha1: manifestSha1 };
        let manifestR;
        try {
            manifestR = await exports.loadFileAsync(node, manifestFileKey, { fromNode, label: `${opts.label} manifest` });
        }
        catch (err) {
            console.warn(`Manifest not found: ${manifestSha1}`);
            ret.producer().error(err);
            return ret;
        }
        if (!manifestR.found) {
            throw Error('Unexpected... loadFileAsync should have thrown an error if not found');
        }
        const manifestDataStream = await node.kacheryStorageManager().getFileReadStream(manifestFileKey);
        const manifestJson = (await manifestDataStream.allData()).toString();
        const manifest = JSON.parse(manifestJson);
        if (!core_1.isFileManifest(manifest)) {
            ret.producer().error(Error('Invalid manifest file'));
            return ret;
        }
        if (!util_1.sha1MatchesFileKey({ sha1: manifest.sha1, fileKey })) {
            ret.producer().error(Error(`Manifest sha1 does not match file key: ${manifest.sha1}`));
            return ret;
        }
        let chunkDataStreams = [];
        const _calculateTotalBytesLoaded = () => {
            let bytesLoaded = 0;
            chunkDataStreams.forEach(ds => {
                bytesLoaded += core_1.byteCountToNumber(ds.bytesLoaded());
            });
            return core_1.byteCount(bytesLoaded);
        };
        const _updateProgressForManifestLoad = () => {
            // this can be made more efficient - don't need to loop through all in-progress files every time
            const bytesLoaded = _calculateTotalBytesLoaded();
            ret.producer().reportBytesLoaded(bytesLoaded);
        };
        const _concatenateChunks = async () => {
            const chunkSha1s = manifest.chunks.map(c => c.sha1);
            await node.kacheryStorageManager().concatenateChunksAndStoreResult(manifest.sha1, chunkSha1s);
            ret.producer().end();
        };
        const _cancelAllChunkDataStreams = () => {
            chunkDataStreams.forEach((ds) => {
                ds.cancel();
            });
            chunkDataStreams = []; // do this so we don't cancel things twice (although it would not be a big deal)
        };
        ret.onError(() => {
            // this will get called if ret is cancelled or if there is another error
            _cancelAllChunkDataStreams();
        });
        (async () => {
            let timer = core_1.nowTimestamp();
            // this happens after ret is returned
            ret.producer().start(manifest.size);
            let numComplete = 0;
            let errored = false;
            try {
                await asyncLoop(manifest.chunks, async (chunk, chunkIndex) => {
                    if (errored)
                        return;
                    const chunkFileKey = {
                        sha1: chunk.sha1,
                        chunkOf: {
                            fileKey: {
                                sha1: manifest.sha1
                            },
                            startByte: chunk.start,
                            endByte: chunk.end
                        }
                    };
                    console.info(`${opts.label}: Handling chunk ${chunkIndex} of ${manifest.chunks.length}`);
                    const label0 = `${opts.label} ch ${chunkIndex}`;
                    const ds = await exports.loadFile(node, chunkFileKey, { fromNode: opts.fromNode, label: label0, _numRetries: 2 });
                    chunkDataStreams.push(ds);
                    return new Promise((resolve, reject) => {
                        ds.onError(err => {
                            errored = true;
                            reject(err);
                        });
                        ds.onProgress((progress) => {
                            _updateProgressForManifestLoad();
                        });
                        ds.onFinished(() => {
                            numComplete++;
                            if (numComplete === manifest.chunks.length) {
                                console.info(`${opts.label}: Concatenating chunks`);
                                _concatenateChunks().then(() => {
                                    const bytesLoaded = _calculateTotalBytesLoaded();
                                    const elapsedSec = core_1.elapsedSince(entireFileTimestamp) / 1000;
                                    const rate = (core_1.byteCountToNumber(bytesLoaded) / 1e6) / elapsedSec;
                                    console.info(`${opts.label}: Downloaded ${createDownloader_1.formatByteCount(bytesLoaded)} in ${elapsedSec} sec [${rate.toFixed(3)} MiB/sec]`);
                                    ret.producer().end();
                                }).catch((err) => {
                                    reject(err);
                                });
                            }
                            resolve();
                        });
                    });
                }, { numSimultaneous: 5 });
            }
            catch (err) {
                ret.producer().error(err);
                return;
            }
        })();
        return ret;
    };
    const loadFileWithoutManifest = async () => {
        let timer = core_1.nowTimestamp();
        await node.downloadOptimizer().waitForReady();
        const ret = new DataStreamy_1.default();
        let fileSize = fileKey.chunkOf ? core_1.byteCount(core_1.byteCountToNumber(fileKey.chunkOf.endByte) - core_1.byteCountToNumber(fileKey.chunkOf.startByte)) : null;
        const numRetries = opts._numRetries === undefined ? 0 : opts._numRetries;
        const task = node.downloadOptimizer().createTask(fileKey, fileSize, opts.label, { fromNode: opts.fromNode, numRetries });
        task.onError(err => {
            ret.producer().error(err);
        });
        task.onProgress((progress) => {
            ret.producer().reportBytesLoaded(progress.bytesLoaded);
        });
        task.onFinished(() => {
            ret.producer().end();
        });
        ret.onError(() => {
            // this gets called on cancel or on any other error
            task.cancel(); // cancel the download
        });
        return ret;
    };
    if (fileKey.manifestSha1) {
        return await loadFileWithManifest();
    }
    else {
        return await loadFileWithoutManifest();
    }
};
exports.loadFile = loadFile;
//# sourceMappingURL=loadFile.js.map