"use strict";
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _db, _initializing, _initialized, _databaseLocked, _onDatabaseLockReleasedCallbacks;
Object.defineProperty(exports, "__esModule", { value: true });
const sqlite_1 = require("sqlite");
const sqlite3_1 = __importDefault(require("sqlite3"));
const crypto_util_1 = require("../common/crypto_util");
const util_1 = require("../common/util");
const core_1 = require("../interfaces/core");
const fs_1 = __importDefault(require("fs"));
class MutableDatabase {
    constructor(databasePath) {
        this.databasePath = databasePath;
        _db.set(this, null);
        _initializing.set(this, false);
        _initialized.set(this, false);
        _databaseLocked.set(this, false);
        _onDatabaseLockReleasedCallbacks.set(this, []);
    }
    async _initialize() {
        if (__classPrivateFieldGet(this, _initialized))
            return;
        while (__classPrivateFieldGet(this, _initializing)) {
            await util_1.sleepMsec(core_1.unscaledDurationMsec(100));
        }
        if (__classPrivateFieldGet(this, _initialized))
            return;
        __classPrivateFieldSet(this, _initializing, true);
        const createTables = async (db) => {
            await db.run(`
                CREATE TABLE IF NOT EXISTS mutables (
                    sha1 TEXT PRIMARY KEY NOT NULL,
                    key TEXT NOT NULL,
                    value TEXT NOT NULL
                ) WITHOUT ROWID;
            `);
        };
        const db = await this._openDatabase();
        try {
            await db.run(`PRAGMA foreign_keys = ON`);
            await createTables(db);
        }
        finally {
            await this._closeDatabase();
        }
        fs_1.default.chmodSync(this.databasePath.toString(), fs_1.default.constants.S_IRUSR | fs_1.default.constants.S_IWUSR);
        __classPrivateFieldSet(this, _initialized, true);
        __classPrivateFieldSet(this, _initializing, false);
    }
    async _acquireDatabaseLock() {
        if (!__classPrivateFieldGet(this, _databaseLocked)) {
            __classPrivateFieldSet(this, _databaseLocked, true);
            return;
        }
        await new Promise((resolve) => {
            __classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks).push(() => {
                __classPrivateFieldSet(this, _databaseLocked, true);
                resolve();
            });
        });
    }
    _releaseDatabaseLock() {
        if (!__classPrivateFieldGet(this, _databaseLocked)) {
            throw Error('Unexpected: (mutable db) cannot release database lock when it is not locked');
        }
        __classPrivateFieldSet(this, _databaseLocked, false);
        if (__classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks).length > 0) {
            const cb = __classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks)[0];
            __classPrivateFieldSet(this, _onDatabaseLockReleasedCallbacks, __classPrivateFieldGet(this, _onDatabaseLockReleasedCallbacks).slice(1));
            cb();
        }
    }
    async _openDatabase() {
        await this._acquireDatabaseLock();
        __classPrivateFieldSet(this, _db, await sqlite_1.open({ filename: this.databasePath.toString(), driver: sqlite3_1.default.Database }));
        return __classPrivateFieldGet(this, _db);
    }
    async _closeDatabase() {
        if (!__classPrivateFieldGet(this, _db))
            throw Error('Unexpected in closeDatabase: (mutable db) database is null');
        await __classPrivateFieldGet(this, _db).close();
        this._releaseDatabaseLock();
    }
    async set(sha1, record) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            await db.run('BEGIN TRANSACTION');
            await db.run(`
                INSERT OR REPLACE INTO mutables (sha1, key, value) VALUES ($sha1, $key, $value)
            `, {
                '$sha1': sha1.toString(),
                '$key': crypto_util_1.JSONStringifyDeterministic(record.key),
                '$value': crypto_util_1.JSONStringifyDeterministic(record.value)
            });
            await db.run('COMMIT');
        }
        finally {
            await this._closeDatabase();
        }
    }
    async delete(sha1) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            await db.run('BEGIN TRANSACTION');
            await db.run(`
                DELETE FROM mutables WHERE sha1 = $sha1
            `, {
                '$sha1': sha1.toString()
            });
            await db.run('COMMIT');
        }
        finally {
            await this._closeDatabase();
        }
    }
    async has(sha1) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            const row = await db.get(`
                SELECT sha1 FROM mutables WHERE sha1 = $sha1
            `, {
                '$sha1': sha1.toString()
            });
            if (row) {
                if (row.sha1 === sha1.toString()) {
                    return true;
                }
                else {
                    // not expected, but checking anyway
                    throw Error('Unexpected: mismatch in sha1');
                }
            }
            else {
                return false;
            }
        }
        finally {
            await this._closeDatabase();
        }
    }
    async get(sha1) {
        await this._initialize();
        const db = await this._openDatabase();
        try {
            const rows = await db.all(`
                SELECT key, value FROM mutables WHERE sha1 = $sha1
            `, {
                '$sha1': sha1.toString()
            });
            if (!rows) {
                throw Error('Unexpected: (mutable db) rows undefined');
            }
            if (rows.length == 0)
                return undefined;
            if (rows.length > 1) {
                throw Error('Unexpected: (mutable db) more than one row found for primary key');
            }
            const row = rows[0];
            const rec = {
                key: JSON.parse(row.key),
                value: JSON.parse(row.value)
            };
            return rec;
        }
        finally {
            await this._closeDatabase();
        }
    }
}
_db = new WeakMap(), _initializing = new WeakMap(), _initialized = new WeakMap(), _databaseLocked = new WeakMap(), _onDatabaseLockReleasedCallbacks = new WeakMap();
exports.default = MutableDatabase;
//# sourceMappingURL=MutableDatabase.js.map