"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var _node, _remoteNodeManager, _halted, _printedWarnings;
Object.defineProperty(exports, "__esModule", { value: true });
const action_1 = require("../common/action");
const constants_1 = require("../common/constants");
const crypto_util_1 = require("../common/crypto_util");
const util_1 = require("../common/util");
const core_1 = require("../interfaces/core");
const protocolVersion_1 = require("../protocolVersion");
const PublicApiServer_1 = require("../services/PublicApiServer");
class BootstrapService {
    constructor(node, opts) {
        this.opts = opts;
        _node.set(this, void 0);
        _remoteNodeManager.set(this, void 0);
        _halted.set(this, false);
        _printedWarnings.set(this, new Set());
        __classPrivateFieldSet(this, _node, node);
        __classPrivateFieldSet(this, _remoteNodeManager, node.remoteNodeManager());
        this._start();
    }
    stop() {
        __classPrivateFieldSet(this, _halted, true);
    }
    async _probeBootstrapNode(address) {
        const response = await __classPrivateFieldGet(this, _node).externalInterface().httpPostJson(address, core_1.urlPath('/probe'), {}, { timeoutMsec: constants_1.TIMEOUTS.defaultRequest });
        if (!PublicApiServer_1.isPublicApiProbeResponse(response)) {
            if ((response.protocolVersion) && (response.protocolVersion !== protocolVersion_1.protocolVersion().toString())) {
                const warningCode = crypto_util_1.JSONStringifyDeterministic(address);
                if (!__classPrivateFieldGet(this, _printedWarnings).has(warningCode)) {
                    __classPrivateFieldGet(this, _printedWarnings).add(warningCode);
                    console.warn('############################');
                    console.warn(`Protocol version does not match with bootstrap node ${address.hostName}:${address.port}: ${protocolVersion_1.protocolVersion()} <> ${response.protocolVersion}`);
                    console.warn('You may need to update one or both of these nodes to the latest version of kachery-p2p');
                    console.warn('');
                }
                return;
            }
            throw Error('Invalid probe response from bootstrap node.');
        }
        const { nodeId: remoteNodeId, isBootstrapNode, webSocketAddress } = response;
        if (!isBootstrapNode) {
            throw Error('isBootstrapNode is false in probe response from bootstrap node.');
        }
        const publicUdpSocketAddress = response.publicUdpSocketAddress;
        __classPrivateFieldGet(this, _remoteNodeManager).setBootstrapNode(remoteNodeId, address, webSocketAddress, publicUdpSocketAddress);
    }
    async _start() {
        const timestampStarted = core_1.nowTimestamp();
        await util_1.sleepMsecNum(2); // important for tests
        // probe the bootstrap nodes periodically
        while (true) {
            if (__classPrivateFieldGet(this, _halted))
                return;
            const startingUp = core_1.elapsedSince(timestampStarted) < core_1.durationMsecToNumber(core_1.scaledDurationMsec(10000));
            const probeIntervalMsec = startingUp ? core_1.scaledDurationMsec(1000) : this.opts.probeIntervalMsec;
            const allBootstrapAddresses = [];
            const channelConfigUrls = __classPrivateFieldGet(this, _node).joinedChannelConfigUrls();
            for (let channelConfigUrl of channelConfigUrls) {
                const channelConfig = await __classPrivateFieldGet(this, _node).getChannelConfig(channelConfigUrl);
                if (channelConfig) {
                    for (let address of channelConfig.bootstrapAddresses) {
                        allBootstrapAddresses.push(address);
                    }
                }
            }
            const uniqueBootstrapAddresses = getUniqueAddresses(allBootstrapAddresses);
            for (let address of uniqueBootstrapAddresses) {
                /////////////////////////////////////////////////////////////////////////
                await action_1.action('probeBootstrapNode', { context: 'BootstrapService', address }, async () => {
                    await this._probeBootstrapNode(address);
                }, async (err) => {
                    console.warn(`Problem probing bootstrap node ${address} (${err.message})`);
                });
                /////////////////////////////////////////////////////////////////////////
            }
            await util_1.sleepMsec(probeIntervalMsec, () => { return !__classPrivateFieldGet(this, _halted); });
        }
    }
}
exports.default = BootstrapService;
_node = new WeakMap(), _remoteNodeManager = new WeakMap(), _halted = new WeakMap(), _printedWarnings = new WeakMap();
const getUniqueAddresses = (addresses) => {
    const used = new Set();
    const ret = [];
    addresses.forEach(a => {
        const code = crypto_util_1.JSONStringifyDeterministic(a);
        if (!used.has(code)) {
            ret.push(a);
            used.add(code);
        }
    });
    return ret;
};
//# sourceMappingURL=BootstrapService.js.map