"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _urisToMirror, _halted, _currentFileIndex, _mirroredFileUris, _mirrorManager, _node, _halted_1, _mirrorSourceManager;
Object.defineProperty(exports, "__esModule", { value: true });
const GarbageMap_1 = __importDefault(require("../common/GarbageMap"));
const util_1 = require("../common/util");
const core_1 = require("../interfaces/core");
const loadFile_1 = require("../loadFile");
const createFileKey = (sha1, query) => {
    const fileKey = { sha1 };
    if ('manifest' in query) {
        fileKey['manifestSha1'] = query.manifest;
    }
    if ('chunkOf' in query) {
        const v = query['chunkOf'].split('~');
        if (v.length !== 3) {
            throw Error(`Unexpected chunkOf in URI query.`);
        }
        fileKey['chunkOf'] = {
            fileKey: { sha1: v[0] },
            startByte: core_1.byteCount(parseInt(v[1])),
            endByte: core_1.byteCount(parseInt(v[2]))
        };
    }
    return fileKey;
};
function parseQueryString(queryString) {
    var query = {};
    var pairs = (queryString[0] === '?' ? queryString.substr(1) : queryString).split('&');
    for (var i = 0; i < pairs.length; i++) {
        var pair = pairs[i].split('=');
        query[decodeURIComponent(pair[0])] = decodeURIComponent(pair[1] || '');
    }
    return query;
}
const parseKacheryUri = (uri) => {
    const listA = uri.split('?');
    const query = (listA.length > 1) ? parseQueryString(listA[1]) : {};
    const list0 = listA[0].split('/');
    const protocol = list0[0].replace(':', '');
    let hash0 = list0[2];
    if (hash0.includes('.')) {
        hash0 = hash0.split('.')[0];
    }
    const additional_path = list0.slice(3).join('/');
    let algorithm = null;
    for (let alg of ['sha1', 'md5', 'key']) {
        if (protocol.startsWith(alg))
            algorithm = alg;
    }
    if (!algorithm) {
        throw Error(`Unexpected protocol in kachery uri: ${protocol}`);
    }
    return {
        protocol, algorithm, hash0, additional_path, query
    };
};
const fileKeyFromUri = (uri) => {
    const { protocol, algorithm, hash0, additional_path, query } = parseKacheryUri(uri);
    if (algorithm !== 'sha1') {
        throw Error(`Unsupported algorith: ${algorithm}`);
    }
    return createFileKey(hash0, query);
};
class MirrorManager {
    constructor(node) {
        this.node = node;
        _urisToMirror.set(this, []);
        _halted.set(this, false);
        _currentFileIndex.set(this, 0);
        _mirroredFileUris.set(this, new GarbageMap_1.default(core_1.scaledDurationMsec(1000 * 60 * 60)));
        this._start();
    }
    setUrisToMirror(x) {
        __classPrivateFieldSet(this, _urisToMirror, x);
    }
    halt() {
    }
    async _mirrorFile(uri) {
        let fileKey;
        fileKey = fileKeyFromUri(uri);
        const hasLocalFile = await this.node.kacheryStorageManager().hasLocalFile(fileKey);
        if (hasLocalFile)
            return;
        console.info(`Mirroring file: ${uri}`);
        return new Promise((resolve, reject) => {
            loadFile_1.loadFile(this.node, fileKey, { fromNode: null, label: `Mirror ${uri}` }).then((x) => {
                x.onFinished(() => {
                    __classPrivateFieldGet(this, _mirroredFileUris).set(uri, true);
                    resolve();
                });
                x.onError((err) => {
                    console.info(`Error mirroring file (${err.message}): ${uri}`);
                    resolve();
                });
                x.onProgress((p) => {
                });
            }).catch((err) => {
                console.info(`Problem mirroring file (${err.message}): ${uri}`);
                resolve();
            });
        });
    }
    async _start() {
        while (true) {
            if (__classPrivateFieldGet(this, _halted))
                return;
            if (__classPrivateFieldGet(this, _currentFileIndex) < __classPrivateFieldGet(this, _urisToMirror).length) {
                const uri = __classPrivateFieldGet(this, _urisToMirror)[__classPrivateFieldGet(this, _currentFileIndex)];
                if (!__classPrivateFieldGet(this, _mirroredFileUris).get(uri)) {
                    try {
                        await this._mirrorFile(uri);
                    }
                    catch (err) {
                        console.warn(`Problem mirroring file (${err.message}): ${uri}`);
                    }
                }
                __classPrivateFieldSet(this, _currentFileIndex, +__classPrivateFieldGet(this, _currentFileIndex) + 1);
            }
            else {
                // wait a bit before starting again
                await util_1.sleepMsec(core_1.scaledDurationMsec(1000 * 120), () => { return !__classPrivateFieldGet(this, _halted); });
                __classPrivateFieldSet(this, _currentFileIndex, 0);
            }
            // duty cycle
            await util_1.sleepMsec(core_1.scaledDurationMsec(30), () => { return !__classPrivateFieldGet(this, _halted); });
        }
    }
}
_urisToMirror = new WeakMap(), _halted = new WeakMap(), _currentFileIndex = new WeakMap(), _mirroredFileUris = new WeakMap();
class MirrorSourceManager {
    constructor(node) {
        this.node = node;
        _mirrorManager.set(this, void 0);
        __classPrivateFieldSet(this, _mirrorManager, new MirrorManager(node));
    }
    async update() {
        const urisToMirror = [];
        const urisToMirrorSet = new Set();
        for (let ms of this.node.mirrorSources()) {
            let uris;
            try {
                uris = await this._getUrisToMirrorFromMirrorSource(ms);
            }
            catch (err) {
                console.warn(`Unable to get uris to mirror (${err.message}): ${ms.uri}`);
                uris = [];
            }
            for (let uri of uris) {
                if (!urisToMirrorSet.has(uri)) {
                    urisToMirror.push(uri);
                    urisToMirrorSet.add(uri);
                }
            }
        }
        __classPrivateFieldGet(this, _mirrorManager).setUrisToMirror(urisToMirror);
    }
    async _getUrisToMirrorFromMirrorSource(x) {
        const fileKey = fileKeyFromUri(x.uri);
        const { found, size } = await loadFile_1.loadFileAsync(this.node, fileKey, { fromNode: null, label: `Mirror source: ${x.uri}` });
        if (!found) {
            throw Error('Unable to find file.');
        }
        const mirrorSourceDataStream = await this.node.kacheryStorageManager().getFileReadStream(fileKey);
        const mirrorSourceText = (await mirrorSourceDataStream.allData()).toString();
        const mirrorSourceContent = JSON.parse(mirrorSourceText);
        return this._getUrisToMirrorFromMirrorSourceContent(mirrorSourceContent);
    }
    _getUrisToMirrorFromMirrorSourceContent(content) {
        if (!content)
            return [];
        if (typeof (content) == 'string') {
            if (content.startsWith('sha1://'))
                return [content];
            else
                return [];
        }
        else if (typeof (content) == 'object') {
            const ret = [];
            if (Array.isArray(content)) {
                for (let a of content) {
                    const b = this._getUrisToMirrorFromMirrorSourceContent(a);
                    for (let uri of b)
                        ret.push(uri);
                }
            }
            else {
                for (let k in content) {
                    const b = this._getUrisToMirrorFromMirrorSourceContent(content[k]);
                    for (let uri of b)
                        ret.push(uri);
                }
            }
            return ret;
        }
        else
            return [];
    }
    halt() {
        __classPrivateFieldGet(this, _mirrorManager).halt();
    }
}
_mirrorManager = new WeakMap();
class MirrorService {
    constructor(node, opts) {
        this.opts = opts;
        _node.set(this, void 0);
        _halted_1.set(this, false);
        _mirrorSourceManager.set(this, void 0);
        __classPrivateFieldSet(this, _node, node);
        __classPrivateFieldSet(this, _mirrorSourceManager, new MirrorSourceManager(node));
        this._start();
    }
    stop() {
        __classPrivateFieldGet(this, _mirrorSourceManager).halt();
        __classPrivateFieldSet(this, _halted_1, true);
    }
    async _start() {
        const intervalMsec = core_1.scaledDurationMsec(1000 * 60 * 3);
        // wait a bit before starting
        await util_1.sleepMsec(core_1.scaledDurationMsec(1000 * 5), () => { return !__classPrivateFieldGet(this, _halted_1); });
        while (true) {
            if (__classPrivateFieldGet(this, _halted_1))
                return;
            await __classPrivateFieldGet(this, _mirrorSourceManager).update();
            await util_1.sleepMsec(intervalMsec, () => { return !__classPrivateFieldGet(this, _halted_1); });
        }
    }
}
exports.default = MirrorService;
_node = new WeakMap(), _halted_1 = new WeakMap(), _mirrorSourceManager = new WeakMap();
//# sourceMappingURL=MirrorService.js.map