"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const core_1 = require("./interfaces/core");
const KacheryP2PNode_1 = __importDefault(require("./KacheryP2PNode"));
const AnnounceService_1 = __importDefault(require("./services/AnnounceService"));
const BootstrapService_1 = __importDefault(require("./services/BootstrapService"));
const ConfigUpdateService_1 = __importDefault(require("./services/ConfigUpdateService"));
const DaemonApiServer_1 = __importDefault(require("./services/DaemonApiServer"));
const DiscoverService_1 = __importDefault(require("./services/DiscoverService"));
const DisplayStateService_1 = __importDefault(require("./services/DisplayStateService"));
const MulticastService_1 = __importDefault(require("./services/MulticastService"));
const ProxyClientService_1 = __importDefault(require("./services/ProxyClientService"));
const PublicApiServer_1 = __importDefault(require("./services/PublicApiServer"));
const PublicUdpSocketServer_1 = __importDefault(require("./services/PublicUdpSocketServer"));
const PublicWebSocketServer_1 = __importDefault(require("./services/PublicWebSocketServer"));
const MirrorService_1 = __importDefault(require("./services/MirrorService"));
const ClientAuthService_1 = __importDefault(require("./services/ClientAuthService"));
const startDaemon = async (args) => {
    const { configDir, // no longer used, but include for migration
    verbose, hostName, publicUrl, daemonApiPort, httpListenPort, label, externalInterface, opts } = args;
    const kNode = new KacheryP2PNode_1.default({
        configDir,
        verbose,
        hostName,
        publicUrl,
        httpListenPort,
        webSocketListenPort: opts.webSocketListenPort,
        udpSocketPort: opts.udpSocketPort,
        label,
        externalInterface,
        opts: {
            isBootstrapNode: opts.isBootstrap,
            multicastUdpAddress: opts.services.multicast ? opts.multicastUdpAddress : null,
            getDefects: args.getDefects
        }
    });
    if ((opts.services.multicast) && (!opts.services.udpSocket)) {
        throw Error('You must enable the udpSocket service when the multicast service is enabled');
    }
    if ((opts.services.udpSocket) && (opts.udpSocketPort === null)) {
        throw Error('Missing udp socket port');
    }
    if ((opts.services.webSocketServer) && (opts.webSocketListenPort === null)) {
        throw Error('Missing web socket port');
    }
    // Start the daemon http server
    const daemonApiServer = new DaemonApiServer_1.default(kNode, { verbose });
    if (opts.services.daemonServer && (daemonApiPort !== null)) {
        await daemonApiServer.listen(daemonApiPort);
        console.info(`Daemon http server listening on port ${daemonApiPort}`);
    }
    // Start the websocket server
    let publicWebSocketServer = null;
    if (opts.services.webSocketServer && opts.webSocketListenPort) {
        publicWebSocketServer = new PublicWebSocketServer_1.default(kNode, { verbose });
        await publicWebSocketServer.startListening(opts.webSocketListenPort);
        console.info(`Websocket server listening on port ${opts.webSocketListenPort}`);
    }
    // Start the udp socket server
    let publicUdpSocketServer = null;
    if (opts.services.udpSocket && opts.udpSocketPort) {
        publicUdpSocketServer = new PublicUdpSocketServer_1.default(kNode, opts.firewalled);
        await publicUdpSocketServer.startListening(opts.udpSocketPort);
        kNode.setPublicUdpSocketServer(publicUdpSocketServer);
        console.info(`Udp socket server listening on port ${opts.udpSocketPort}`);
    }
    // start the other services
    const announceService = opts.services.announce ? new AnnounceService_1.default(kNode, {
        announceBootstrapIntervalMsec: core_1.scaledDurationMsec(21000),
        announceToIndividualNodeIntervalMsec: core_1.scaledDurationMsec(2000)
    }) : null;
    const discoverService = opts.services.discover ? new DiscoverService_1.default(kNode, {
        discoverBootstrapIntervalMsec: core_1.scaledDurationMsec(30000),
        discoverIndividualNodeIntervalMsec: core_1.scaledDurationMsec(1500)
    }) : null;
    const bootstrapService = opts.services.bootstrap ? new BootstrapService_1.default(kNode, {
        probeIntervalMsec: core_1.scaledDurationMsec(15000)
    }) : null;
    const proxyClientService = opts.services.proxyClient ? new ProxyClientService_1.default(kNode, {
        intervalMsec: core_1.scaledDurationMsec(3000)
    }) : null;
    const configUpdateService = opts.services.configUpdate ? new ConfigUpdateService_1.default(kNode, {
        intervalMsec: core_1.scaledDurationMsec(5000),
        staticConfigPathOrUrl: opts.staticConfigPathOrUrl
    }) : null;
    let multicastService = (opts.services.multicast && (opts.multicastUdpAddress !== null)) ? new MulticastService_1.default(kNode, {
        intervalMsec: core_1.scaledDurationMsec(12000),
        multicastAddress: opts.multicastUdpAddress
    }) : null;
    let displayService = opts.services.display ? new DisplayStateService_1.default(kNode, {
        daemonApiPort
    }) : null;
    const mirrorService = opts.services.mirror ? new MirrorService_1.default(kNode, {
        intervalMsec: core_1.scaledDurationMsec(120000)
    }) : null;
    const clientAuthService = opts.services.clientAuth ? new ClientAuthService_1.default(kNode, {
        clientAuthGroup: opts.authGroup ? opts.authGroup : null
    }) : null;
    // Start the public http server
    const publicApiServer = new PublicApiServer_1.default(kNode, { verbose });
    if (opts.services.httpServer && httpListenPort) {
        publicApiServer.listen(httpListenPort);
        console.info(`Public http server listening on port ${httpListenPort}`);
    }
    const _stop = () => {
        announceService && announceService.stop();
        discoverService && discoverService.stop();
        bootstrapService && bootstrapService.stop();
        proxyClientService && proxyClientService.stop();
        configUpdateService && configUpdateService.stop();
        multicastService && multicastService.stop();
        displayService && displayService.stop();
        mirrorService && mirrorService.stop();
        clientAuthService && clientAuthService.stop();
        // wait a bit after stopping services before cleaning up the rest (for clean exit of services)
        setTimeout(() => {
            daemonApiServer && daemonApiServer.stop();
            publicApiServer && publicApiServer.stop();
            setTimeout(() => {
                publicWebSocketServer && publicWebSocketServer.stop();
                publicUdpSocketServer && publicUdpSocketServer.stop();
                kNode.cleanup();
            }, 20);
        }, 20);
    };
    return {
        daemonApiServer,
        publicApiServer,
        publicWebSocketServer,
        publicUdpSocketServer,
        announceService,
        discoverService,
        bootstrapService,
        proxyClientService,
        configUpdateService,
        multicastService,
        displayService,
        mirrorService,
        clientAuthService,
        node: kNode,
        stop: _stop
    };
};
exports.default = startDaemon;
//# sourceMappingURL=startDaemon.js.map